#!/usr/bin/perl
#==============================================================================
#
#               (c) Paul Nesfield, 2005 - 2009
#
# TITLE:        Diabetes Report
# VERSION:      1.5.3
# DATE:         23 March 2009
# AUTHOR:       Paul Nesfield
#
# SUMMARY
#
# A perl script to extract diabetes information and produce a HTML format
# report. Modified from my generic PERL script, Diabetes Report v0.1.
# 
# The supported format requires a subroutine to process the logs, this 
# includes a definition of where the parameters are in the log file and 
# routines to extract them.
# 
#==============================================================================

use Getopt::Std;
use Time::Local;

#==============================================================================
#
# DEFINE VARIABLES - START
#
#==============================================================================

#Define clean variables;
undef %Parameters;
undef %Glucose;
undef %Insulin;
undef %InsType;
undef %Carbs;
undef %InsulinB;
undef %InsBType;
undef %Medicine;
undef %MedName;
undef %Ketones;
undef %KetValue;
undef %Eventname;
undef %Notes;
undef %Weight;
undef %Height;
undef %HbA1c;
undef %Notable;
undef %Period;
undef %Date;
undef %Day;
undef %Time;
undef %Timestamp;
undef @Records;
undef %Periods;
undef $Events;
undef %LocStr;
undef %BMI;


getopts('cdD:E:F:ghH:iI:lL:mnN:sS:T:U:W:X:Y:Z:');

# Parameters:
# c: include carbs flag
# d: include detailed report flag
# D: the data, one event per line
# E: report end date in YYYYMMDD format
# F: format for dates = UK, US
# g: include graph report flag
# h: include health flag
# H: Height units for the report
# i: include insulin flag
# I: Weight units for the report
# l: include logbook report flag
# L: path to localisation file
# m: include medicines flag
# n: include notes report flag
# N: user name to include in report header
# s: include summary report flag
# S: report start date in YYYYMMDD format
# T: report time format 24HR/12HR
# U: glucose units of the report
# W: critical low glucose level
# X: target low glucose level
# Y: target high glucose level
# Z: critical high glucose level


#==========================================
# LOCALISABLE VARIABLES - PLEASE MODIFY
#==========================================
#Retrieve the logbook data direct from the Mac OS X defaults system
undef $locdatastr;
open(LOC, $opt_L);
while (<LOC>) {
    $locdatastr .= $_."\n";
}
close(LOC);
$locdatastr =~ s/\c@//g;
$locdatastr =~ s/\r/\n/g;
$locdatastr =~ s/\f/\n/g;
$locdatastr =~ s/\cM/\n/g;
$locdatastr =~ s/.*\/\*.*\*\///g;
(@locdata) = split /"\s*;\s*\n/, $locdatastr;
  
foreach (@locdata) {
  $locstr=$_;
  $locstr=~ s/^\s*"//;
  ($key, $value) = split /"\s*=\s*"/, $locstr;
  $LocStr{$key}=$value if ($key);
}

%DaysOfWeek=(
  "Sun"   => $LocStr{"Sun"},
  "Mon"   => $LocStr{"Mon"},
  "Tue"   => $LocStr{"Tue"},
  "Wed"   => $LocStr{"Wed"},
  "Thu"   => $LocStr{"Thu"},
  "Fri"   => $LocStr{"Fri"},
  "Sat"   => $LocStr{"Sat"}
);

@TimePeriods=(
  $LocStr{"Before Breakfast"},
  $LocStr{"After Breakfast"},
  $LocStr{"Before Lunch"},
  $LocStr{"After Lunch"},
  $LocStr{"Before Dinner"},
  $LocStr{"After Dinner"},
  $LocStr{"Before Bedtime"},
  $LocStr{"Night"}
);

%TimePeriods=(
  $LocStr{"Before Breakfast"} => "NO",
  $LocStr{"After Breakfast"}  => "NO",
  $LocStr{"Before Lunch"}     => "NO",
  $LocStr{"After Lunch"}      => "NO",
  $LocStr{"Before Dinner"}    => "NO",
  $LocStr{"After Dinner"}     => "NO",
  $LocStr{"Before Bedtime"}   => "NO",
  $LocStr{"Night"}            => "NO"
);
        
%Time12Hr=(
  "AM"  => "AM",
  "PM"  => "PM"
);

#Which localisation
$LocLoc=$LocStr{"Localisation"};

#Report Header texts
$LocReportHeader=$LocStr{"Diabetes Report"};
$LocFor=$LocStr{" for "};

#Report Footer texts
$LocReportFooter=$LocStr{"Report produced using"};

#Header Notes section
$LocNotes=$LocStr{"Notes:"};
$LocNotes1=$LocStr{"Glucose units are"};
$LocNotes2=$LocStr{"Notable/flagged events are highlighted like this."};
$LocNotes3=$LocStr{"Insulins, carbohydrates and medicines are listed as follows:"};
$LocInsulinA=$LocStr{"Insulin A"};
$LocInsulinB=$LocStr{"Insulin B"};
$LocCarbs=$LocStr{"Carbs."};
$LocMedicine=$LocStr{"Medicine"};
$LocNotes4=$LocStr{"Carbs. are in the user's unit of preference, e.g. portion or grammes."};
$LocWeight=$LocStr{"Weight"};
$LocHeight=$LocStr{"Height"};
$LocHbA1c=$LocStr{"HbA1c"};
$LocBMI=$LocStr{"BMI"};

#Header Level Indicator section
$LocGluc=$LocStr{"Glucose Level Indicator Information:"};
$LocGluc1a=$LocStr{"Very high glucose level set"};
$LocGluc1b=$LocStr{"above"};
$LocGluc2a=$LocStr{"High glucose level set"};
$LocGluc2b=$LocStr{"above"};
$LocGluc3a=$LocStr{"Target glucose level set"};
$LocGluc3c=$LocStr{"between"};
$LocGluc3c=$LocStr{"and"};
$LocGluc4a=$LocStr{"Low glucose level set"};
$LocGluc4b=$LocStr{"below"};
$LocGluc5a=$LocStr{"Very low glucose level set"};
$LocGluc5b=$LocStr{"below"};

#Graph report
$LocGraphHeader=$LocStr{"Glucose Events"};
$LocGraphHbA1cHeader=$LocStr{"HbA1c History"};
$LocGraphWeightHeader=$LocStr{"Weight Record"};
$LocGraphHeightHeader=$LocStr{"Height Record"};
$LocGraphBMIHeader=$LocStr{"BMI Report"};
$LocGraph0=$LocStr{"Legend:"};
$LocGraph1=$LocStr{"Critical high/low"};
$LocGraph2=$LocStr{"Target high/low"};
$LocGraph3=$LocStr{"Average glucose for period"};
$LocGraph4=$LocStr{"Day separater"};
$LocGraph5=$LocStr{"* last 300 readings"};
$LocGraph6=$LocStr{"Ideal HbA1c"};
$LocGraph7=$LocStr{"* last 150 readings"};

#Graph Summary report
$LocGraphSummaryHeader=$LocStr{"Average Glucose by Time Period"};
$LocGraphSummary0=$LocStr{"The graph shows the average of the glucose readings (diamond) with error bars showing the Standard Deviation [the average amount that readings vary from the calculated average glucose]"};
$LocGraphSummary1=$LocStr{"Deviation"};

#Notes report
$LocNotesHeader=$LocStr{"Notes"};
$LocNotesCol1=$LocStr{"Date"};
$LocNotesCol2=$LocStr{"Time"};
$LocNotesCol3=$LocStr{"Event"};
$LocNotesCol4=$LocStr{"Details"};
$LocWeekStarting=$LocStr{"Week starting"};

#Detailed/Summary report
$LocSummaryHeader=$LocStr{"Summary for"};
$LocDetailedHeader=$LocStr{"Detailed Report"};
$LocReportCol1=$LocStr{"Date"};
$LocReportCol2=$LocStr{"Period"};
$LocReportCol3=$LocStr{"Time"};
$LocReportCol4=$LocStr{"Glucose"};
$LocReportCol5=$LocStr{"Insulin & Meds"};
$LocReportCol5a=$LocStr{"Carbs & Health"};
$LocReportCol6=$LocStr{"Notes"};
$LocReportNotes1=$LocStr{"Ketones: +ve!"};
$LocReportNotes2=$LocStr{"Ketones: -ve"};
$LocReportNotes3=$LocStr{"Event"};

$LocAverage=$LocStr{"Average"};
$LocMinimum=$LocStr{"Minimum"};
$LocMaximum=$LocStr{"Maximum"};

#Logbook report
$LocLogbookHeader=$LocStr{"Glucose Logbook"};


#Average Medication report
$LocAvgMedHeader=$LocStr{"Average Medication Summary (Doses)"};
$LocAvgMedCol1=$LocStr{"Medication"};
$LocAvgMedCol2=$LocStr{"DAILY AVERAGE"};


#==========================================
# NORMAL VARIABLES - DO NOT MODIFY
#==========================================


#If desired add patient's name for inclusion in the report header
$PatientName="";
$PatientName="$opt_N" if ($opt_N);

#Set default Glucose units: used to set graph range
$GlucoseUnits="mmol/L";
$GlucoseUnits="$opt_U" if ($opt_U);
if ($GlucoseUnits eq "mmol/L") {
  $GraphYscale="<table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=100% height=100%>";
  $GraphYscale.="  <tr><td valign=top align=right height=15>> 22.0</td></tr>";
  $GraphYscale.="  <tr><td valign=bottom align=right height=10>20.0</td></tr>";
  $GraphYscale.="  <tr><td valign=bottom align=right height=20>18.0</td></tr>";
  $GraphYscale.="  <tr><td valign=bottom align=right height=20>16.0</td></tr>";
  $GraphYscale.="  <tr><td valign=bottom align=right height=20>14.0</td></tr>";
  $GraphYscale.="  <tr><td valign=bottom align=right height=20>12.0</td></tr>";
  $GraphYscale.="  <tr><td valign=bottom align=right height=20>10.0</td></tr>";
  $GraphYscale.="  <tr><td valign=bottom align=right height=20>8.0</td></tr>";
  $GraphYscale.="  <tr><td valign=bottom align=right height=20>6.0</td></tr>";
  $GraphYscale.="  <tr><td valign=bottom align=right height=20>4.0</td></tr>";
  $GraphYscale.="  <tr><td valign=bottom align=right height=20>2.0</td></tr>";
  $GraphYscale.="  <tr><td valign=bottom align=right height=15>0.0</td></tr>";
  $GraphYscale.="</table>";
}
else {
  $GraphYscale="<table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=100% height=100%>";
  $GraphYscale.="  <tr><td valign=top align=right height=15>> 440</td></tr>";
  $GraphYscale.="  <tr><td valign=bottom align=right height=10>400</td></tr>";
  $GraphYscale.="  <tr><td valign=bottom align=right height=20>360</td></tr>";
  $GraphYscale.="  <tr><td valign=bottom align=right height=20>320</td></tr>";
  $GraphYscale.="  <tr><td valign=bottom align=right height=20>280</td></tr>";
  $GraphYscale.="  <tr><td valign=bottom align=right height=20>240</td></tr>";
  $GraphYscale.="  <tr><td valign=bottom align=right height=20>200</td></tr>";
  $GraphYscale.="  <tr><td valign=bottom align=right height=20>160</td></tr>";
  $GraphYscale.="  <tr><td valign=bottom align=right height=20>120</td></tr>";
  $GraphYscale.="  <tr><td valign=bottom align=right height=20>80</td></tr>";
  $GraphYscale.="  <tr><td valign=bottom align=right height=20>40</td></tr>";
  $GraphYscale.="  <tr><td valign=bottom align=right height=15>0.0</td></tr>";
  $GraphYscale.="</table>";
}

#Set default Weight units: used to set graph range
$WeightUnits="Kg";
$WeightUnits="$opt_I" if ($opt_I);

#Set default Glucose units: used to set graph range
$HeightUnits="cm";
$HeightUnits="$opt_H" if ($opt_H);

#Set default BMI units
$BMIUnits="Kg/m&sup2;";

#Set default report colour
$ReportColour="DARKBLUE";
$HeadingColour="WHITE";
$TextColour="BLACK";

#Multiplier for correcting graph scale
$MGperL=18;

#Set default UK or US to have day/month or month/day
$DateFormat="day/month/year";
$DateFormat="day/month/year" if ($opt_F eq "UK");
$DateFormat="month/day/year" if ($opt_F eq "US");
$DateSep="/";
$DateSep="&#046;" if ($LocLoc eq "German");

#Set default 24Hr or 12Hr for displayed time format
$TimeFormat="24HR";
$TimeAdjustWidth=0;
$TimeFormat=$opt_T if ($opt_T);
$TimeAdjustWidth=15 if ($TimeFormat eq "12HR");
$InsAdjustWidth=0;
$InsAdjustWidth=116 if ($opt_i || $opt_m);
$MedsAdjustWidth=0;
$MedsAdjustWidth=94 if ($opt_h || $opt_c);

#Set default reports to include
$NotesReport="NO";
$NotesReport="YES" if ($opt_n);
$GraphReport="NO";
$GraphReport="YES" if ($opt_g);
$DetailedReport="NO";
$DetailedReport="YES" if ($opt_d);
$SummaryReport="NO";
$SummaryReport="YES" if ($opt_s);
$LogbookReport="NO";
$LogbookReport="YES" if ($opt_l);

#================================
#Set for use in selecting colours
#for displaying Glucose data:
# 
# Level                  Colour
#  
# Below 'CriticalLow'  = 'VeryLow'
# --"-- 'TargetLow'    = 'Low'
# --"-- 'TargetHigh'   = 'Target'
# --"-- 'CriticalHigh' = 'High'
# Above 'CriticalHigh' = 'VeryHigh'
#

%GlucoseColors=(
  "Target"     => "Green",
  "Low"        => "Orange",
  "High"       => "Orange",
  "VeryLow"    => "Red",
  "VeryHigh"   => "Red"
);

%GlucoseLevels=(
  "CriticalLow"   => $opt_W,
  "TargetLow"     => $opt_X,
  "TargetHigh"    => $opt_Y,
  "CriticalHigh"  => $opt_Z
  );

# Restrict to set dates
$ReportStartDate=$opt_S if ($opt_S);
$ReportEndDate=$opt_E if ($opt_E);
$StartDate="";
$EndDate="";

#==============================================================================
#
# DEFINE VARIABLES - END
#
#==============================================================================


#==============================================================================
#
# MAIN PROCESS - START
#
#==============================================================================

#======================
# Extract Data
#======================

#New routine for retrieving from the user defaults -faster & avoids limitation
#on size of data that can be parameterised. Loads variable $Events
GetData();

#Split the input data into separate events at the event delimiter {dlbxreturn}
(@Records) = split "{dlbxreturn}" , $Events;

#Process each event in the log
foreach (@Records) {
    
  #Prepare fresh array for the event
  undef @Elements;
    
  #The event data is passed as the current $_
  ParseLogs();
}

#Get time ordered list of timestamps
undef $myHeight;
undef $myWeight;
(@Timestamp)=sort NormalSort (keys %Date);
foreach (@Timestamp) {
  $StartDate=$Date{$_} if (!$StartDate);
  $EndDate=$Date{$_};
  if ($Height{$_}) {
    $myHeight=$Height{$_};
    $myHeight=sprintf("%0.1f", $myHeight * 2.54) if ($HeightUnits ne "cm");
  }
  if ($myHeight && $Weight{$_}) {
    $myWeight=$Weight{$_};
    $myWeight=sprintf("%0.1f", $myWeight * 0.454) if ($WeightUnits ne "Kg");
    $BMI{$_}=sprintf("%0.1f", $myWeight/(($myHeight/100)**2));
  }
}


#======================
# Generate Report
#======================

#Print the report header
PrintReportHeader();

#Sort periods by time
(@Times)=sort NormalSort (values %Periods);
foreach (@TimePeriods) {
  if ($TimePeriods{$_} eq "YES") {
    $NumPeriods=push @Periods, $_;
    $PeriodPosition{$_}=scalar $NumPeriods - 1;
    $TotalPeriods=$NumPeriods;
    $NumPeriods--;
  }
}
foreach (@Times) {
  ($pad,$vPeriod)=split " ", $_, 2;
  $NumPeriods=push @Periods, $vPeriod;
  $PeriodPosition{$vPeriod}=scalar $NumPeriods - 1;
  $TotalPeriods=$NumPeriods;
  $NumPeriods--;
}



#Print the glucose graph
if ($GraphReport eq "YES") {
  PrintGraph();
  PrintGraphSummary();
  if ($opt_h) {
    PrintGraphHbA1c();
    PrintGraphWeight();
    PrintGraphHeight();
    PrintGraphBMI();
  }
}
#Print the glucose graph
PrintNotes() if ($NotesReport eq "YES");


#Print the detailed report
PrintReport() if ($DetailedReport eq "YES");


#Print the summary report for each period, sorted alphabetically
if ($SummaryReport eq "YES") {
  undef $FirstPeriod;
  foreach (@Periods) {
    $ReportPeriod=$_;
    PrintReport();
  }
}


#Print the logbook report
if ($LogbookReport eq "YES") {
  PrintLogbookReport();
}


#Print the average usage summanry report
if (($LogbookReport eq "YES" || $GraphReport eq "YES" || $SummaryReport eq "YES" || $DetailedReport eq "YES") && ($opt_i || $opt_m)) {
  PrintInsMedAvgReport();
}


#Print the report footer
PrintReportFooter();

#==============================================================================
#
# MAIN PROCESS - END
#==============================================================================





#==============================================================================
#
# SUBROUTINES - GET, PARSE and LOAD DATA
#
#==============================================================================

#======================================================================
# Load data from the user defaults.
#======================================================================
sub GetData {

  #Retrieve the logbook data direct from the Mac OS X defaults system
  open(DEFS, "/usr/bin/defaults read com.Nesfield.DiabetesLogbookX EventTableV2|");
  @eventdata = <DEFS>;
  close(DEFS);
  
  #Parse each line and determine if it is a record, start assumption is 'not'
  $recordevent="NO";
  foreach (@eventdata) {
  
    #Strip out all the padding, replace chars with HTML friendly alternatives
    $event=$_;
    $event=~ s/^\s+//;
    $event=~ s/ \n/\n/;
    $event=~ s/^"//;
    $event=~ s/,$//;
    $event=~ s/"$//;
    $event=~ s/\\\\n/<br>/g;
  

    #Events start with a ( line and end with a ) line
    if ($event eq "(\n") {
      $recordevent="YES";
    }
    elsif ($event eq ")\n") {
      $recordevent="STOP";
    }

    #Ignore event start line and start recording event elements,
    #close the event when event stop line detected.
    if ($recordevent eq "YES") {
      $recordevent="SAVE";
    }
    elsif ($recordevent eq "SAVE") {
      $event=~ s/\n//;
      #Add event element delimiter {dlbxtab}
      $Events .= $event."{dlbxtab}";
    }
    elsif ($recordevent eq "STOP") {
      #Add two extra event element delimiters for use during parsing
      #Add end of event delimiter {dlbxreturn}
      $Events .= "{dlbxtab}{dlbxtab}{dlbxreturn}";
      $recordevent="NO";
    }
  }

  #Perform global replace for TAB chars to maintain desired spacing
  $Events=~ s/\\\  /&nbsp;&nbsp;&nbsp;&nbsp;/g;
  $Events=~ s/\\\\t/&nbsp;&nbsp;&nbsp;&nbsp;/g;

  #Perform global replace for quote chars to HTML char code to stop
  #breaking any HTML foramtting usage of quotes
  $Events=~ s/\\\\"/&quot;/g;

  #Perform global replace for unicode chars to HTML char codes
  $Events=~ s/\\\\U(\w{4}?)/&#x$1;/g;


  # !!! All event data is now in the string $Events !!!
  
}


#======================================================================
# Log Parsing and and extraction.
#======================================================================
sub ParseLogs {

  #Process the log entry, passed in with the current $_
  ParseDEFAULT();

  #Load the data into the various Hash tables, using the data now
  #loaded and formated in the @Elements array
  LoadLogs ();
  
}


#======================================================================
#Parse DEFAULT Diabetes Log export files from GetData
#======================================================================
sub ParseDEFAULT {

  #The location of the data in the @Elements array
  #Must set how the LoadLogs sub uses the data.
  #Day element initially holds the date and time. After parsing the Date and
  #Time are move to {24} and {25} and the short dayname is placed in {1}
  %Parameters=(
    "Notable"    =>  0,
    "Date"       =>  24,
    "Time"       =>  25,
    "Period"     =>  2,
    "Glucose"    =>  8,
    "GlucUnits"  =>  9,
    "Insulin"    =>  10,
    "InsType"    =>  11,
    "Carbs"      =>  12,
    "InsulinB"   =>  13,
    "InsBType"   =>  14,
    "Medicine"   =>  15,
    "MedName"    =>  16,
    "Ketones"    =>  17,
    "KetValue"   =>  18,
    "Eventname"  =>  19,
    "Notes"      =>  20,
    "Weight"     =>  21,
    "Height"     =>  22,
    "HbA1c"      =>  23,
    "Day"        =>  1
  );

  #Split log at the {dlbxtab} delimiter marks
  (@Elements) = split "{dlbxtab}";

  #Process date and time to ensure format: YYYY, MM, DD, hh, mm
  ($Date, $Time, $pad) = split " ", $Elements[$Parameters{"Day"}];
  ($Year, $Month, $Day) = split "-", $Date;
  ($Hour, $Minute, $Second) = split ":", $Time;
  
  $Year=scalar 2000 + $Year if ($Year < 1970);
  
  #Get the short day name - needs to be localised for other languages
  $ThisTime=timelocal("","","", scalar $Day, scalar $Month - 1, scalar $Year - 1900);
  $DayName=($DaysOfWeek{"Sun"}, $DaysOfWeek{"Mon"}, $DaysOfWeek{"Tue"}, $DaysOfWeek{"Wed"}, $DaysOfWeek{"Thu"}, $DaysOfWeek{"Fri"}, $DaysOfWeek{"Sat"})[(localtime($ThisTime))[6]] if ($Month && (scalar $Month <= 12));

  #Reload the data into the final locations
  $Elements[$Parameters{"Date"}] = $Day.$DateSep.$Month.$DateSep.$Year if ($DateFormat eq "day/month/year");
  $Elements[$Parameters{"Date"}] = $Month.$DateSep.$Day.$DateSep.$Year if ($DateFormat eq "month/day/year");
  if ($TimeFormat eq "24HR") {
    $Elements[$Parameters{"Time"}] = $Hour.":".$Minute ;
  }
  else {
    $my12Hour=sprintf("%0.0f", $Hour);
    if ($my12Hour >= 12) {
      $my12Hour=$my12Hour - 12 if ($my12Hour > 12);
      $Elements[$Parameters{"Time"}] = $my12Hour.":".$Minute." ".$Time12Hr{"PM"} ;
    }
    else {
      $my12Hour="12" if ($my12Hour == 0);
      $Elements[$Parameters{"Time"}] = $my12Hour.":".$Minute." ".$Time12Hr{"AM"} ;
    }
  }
  $Elements[$Parameters{"Day"}] = $DayName;

}


#==============================================================================
#
# SUBROUTINES - DO NOT MODIFY
#
#==============================================================================


#=========================================
# Load extracted data into the Hash tables
#=========================================
sub LoadLogs {

  my $TempTime;
  my $Weight;
  
  #Build a timestamp to index the Hash tables
  $Timestamp=$Year.$Month.$Day.$Hour.$Minute.$Second;
  $Datestamp=$Year.$Month.$Day;
  $TempTime=$Hour.$Minute;
  chomp $TempTime;

  #Filter required dates, if set
  if (($ReportStartDate && $ReportEndDate) && (($Datestamp < $ReportStartDate) || ($Datestamp > $ReportEndDate))) {
    $Timestamp="";
  }

  #Does the event have any data?
  $HasData="NO";
  $HasData="YES" if ($Elements[$Parameters{"Glucose"}]);
  $HasData="YES" if ($Elements[$Parameters{"Insulin"}]);
  $HasData="YES" if ($Elements[$Parameters{"Carbs"}]);
  $HasData="YES" if ($Elements[$Parameters{"InsulinB"}]);
  $HasData="YES" if ($Elements[$Parameters{"Medicine"}]);
  $HasData="YES" if ($Elements[$Parameters{"Ketones"}]);
  $HasData="YES" if ($Elements[$Parameters{"Notes"}]);
  $HasData="YES" if ($Elements[$Parameters{"Weight"}]);
  $HasData="YES" if ($Elements[$Parameters{"Height"}]);
  $HasData="YES" if ($Elements[$Parameters{"HbA1c"}]);
  
  #If the timestamp is 14 chars then it is a valid
  #date and time, so we add the data to the tables
  if (length $Timestamp == 14 && ($HasData eq "YES")) {
    
    #Load in the data using the @Elements array and the parameter location
    #defined in ParseDEFAULT sub. Only load if the data exists
    
    $Notable{$Timestamp}="!" if ($Elements[$Parameters{"Notable"}]);
    $Date{$Timestamp}=$Elements[$Parameters{"Date"}];
    $Day{$Timestamp}=$Elements[$Parameters{"Day"}];
    $Time{$Timestamp}=$Elements[$Parameters{"Time"}];
    $Period{$Timestamp}=$Elements[$Parameters{"Period"}] if ($Elements[$Parameters{"Period"}]);
    $TimePeriods{$Elements[$Parameters{"Period"}]}="YES" if ($TimePeriods{$Elements[$Parameters{"Period"}]});

    #Let's do this the long way and not get it wrong!
    #Convert glucose readings to the 'master' glucose units and format accordingly
    if (($Elements[$Parameters{"Glucose"}]) && ($Elements[$Parameters{"GlucUnits"}])) {
      if ($Elements[$Parameters{"GlucUnits"}] eq "mmol/L") {
        if ($GlucoseUnits eq "mg/dL") {
          $Glucose{$Timestamp}=sprintf("%0.0f", $Elements[$Parameters{"Glucose"}]*$MGperL);
        }
        else {
          $Glucose{$Timestamp}=sprintf("%0.1f", $Elements[$Parameters{"Glucose"}]);
        }
      }
      elsif ($Elements[$Parameters{"GlucUnits"}] eq "mg/dL") {
        if ($GlucoseUnits eq "mmol/L") {
          $Glucose{$Timestamp}=sprintf("%0.1f", $Elements[$Parameters{"Glucose"}]/$MGperL);
        }
        else {
          $Glucose{$Timestamp}=sprintf("%0.0f", $Elements[$Parameters{"Glucose"}]);
        }
      }
    }
    
  if ($Elements[$Parameters{"Insulin"}]) {
    $Value1=sprintf("%0.1f", $Elements[$Parameters{"Insulin"}]);
    $Value2=sprintf("%0.2f", $Elements[$Parameters{"Insulin"}]);
      $Insulin{$Timestamp}.="<br/>" if ($Insulin{$Timestamp});
    if ($Value1 == $Value2) {
      $Insulin{$Timestamp}.=$Value1;
    }
    else {
      $Insulin{$Timestamp}.=$Value2;
      }
  }
    if ($Elements[$Parameters{"InsType"}]) {
    $InsType{$Timestamp}.="<br/>" if ($InsType{$Timestamp});
    $InsType{$Timestamp}.=$Elements[$Parameters{"InsType"}];
    }
    $Carbs{$Timestamp}=sprintf("%0.1f", $Elements[$Parameters{"Carbs"}]) if ($Elements[$Parameters{"Carbs"}]);
  if ($Elements[$Parameters{"InsulinB"}]) {
    $Value1=sprintf("%0.1f", $Elements[$Parameters{"InsulinB"}]);
    $Value2=sprintf("%0.2f", $Elements[$Parameters{"InsulinB"}]);
        $InsulinB{$Timestamp}.="<br/>" if ($InsulinB{$Timestamp});
    if ($Value1 == $Value2) {
      $InsulinB{$Timestamp}.=$Value1;
    }
    else {
      $InsulinB{$Timestamp}.=$Value2;
    }
  }
    if ($Elements[$Parameters{"InsBType"}]) {
    $IshBType{$Timestamp}.="<br/>" if ($InsBType{$Timestamp});
    $InsBType{$Timestamp}.=$Elements[$Parameters{"InsBType"}];
    }
  
  if ($Elements[$Parameters{"Medicine"}]) {
    $Value1=sprintf("%0.1f", $Elements[$Parameters{"Medicine"}]);
    $Value2=sprintf("%0.2f", $Elements[$Parameters{"Medicine"}]);
        $Medicine{$Timestamp}.="<br/>" if ($Medicine{$Timestamp});
    if ($Value1 == $Value2) {
      $Medicine{$Timestamp}.=$Value1;
    }
    else {
      $Medicine{$Timestamp}.=$Value2;
    }
  }
    if ($Elements[$Parameters{"MedName"}]) {
    $MedName{$Timestamp}.="<br/>" if ($MedName{$Timestamp});
    $MedName{$Timestamp}.=$Elements[$Parameters{"MedName"}];
    }
    $Ketones{$Timestamp}=$Elements[$Parameters{"Ketones"}] if ($Elements[$Parameters{"Ketones"}]);
    $KetValue{$Timestamp}=sprintf("%0.1f", $Elements[$Parameters{"KetValue"}]) if ($Elements[$Parameters{"KetValue"}]);
    $Eventname{$Timestamp}=$Elements[$Parameters{"Eventname"}] if ($Elements[$Parameters{"Eventname"}]);
    $Notes{$Timestamp}=$Elements[$Parameters{"Notes"}] if ($Elements[$Parameters{"Notes"}]);
    if ($Elements[$Parameters{"Weight"}]) {
      $Weight{$Timestamp}=sprintf("%0.1f", $Elements[$Parameters{"Weight"}]);
    }
    if ($Elements[$Parameters{"Height"}]) {
      $Height{$Timestamp}=sprintf("%0.1f", $Elements[$Parameters{"Height"}]);
    }
    $HbA1c{$Timestamp}=sprintf("%0.1f", $Elements[$Parameters{"HbA1c"}]) if ($Elements[$Parameters{"HbA1c"}]);

    #Record the time periods for use in the summary reports
    if ($Elements[$Parameters{"Glucose"}]) {
      if (!$TimePeriods{$Period{$Timestamp}}) {
        ($OldTime,$pad)=split / / , $Periods{$Period{$Timestamp}}, 2;
        if ($OldTime lt $TempTime) {
          #This records the latest time that the period was used. So they can be
          #printed out earliest period first through to latest period
          $Periods{$Period{$Timestamp}}=$TempTime." ".$Period{$Timestamp};
        }
      }
    }
  }
}

#==============================================================================
#
# SUBROUTINES - GET, PARSE and LOAD DATA   --- END
#
#==============================================================================




#==============================================================================
#
# SUBROUTINES - PRINT REPORTS
#
#==============================================================================

#=======================================
# Report Generation Subroutines - Header
#=======================================
sub PrintReportHeader() {

  #Add HTML headers plus Title, report dates, and patient name , if set
  print "<html>
  <head>
    <title>".$LocReportHeader;
    
  print $LocFor.$PatientName if ($PatientName);
  
  print " : ".$StartDate." - ".$EndDate."</title>
    <style type=\"text/css\">
      <!--
        th {bgcolor: ".$ReportColour."; font-family:arial; font-size: 16px; color: ".$HeadingColour."}
        td {bgcolor: white; font-family:arial; font-size: 10px; color: ".$TextColour."}
        -->
    </style>
  </head>
  <body><center>
    <table bgcolor=".$ReportColour." border=0 cellspacing=0 cellpadding=0 width=683>
      <tr>
        <td align=center style=\"font-family:arial; font-size: 18px; color: ".$HeadingColour.";\" width=10%>
          <b>".$LocReportHeader;

  print $LocFor.$PatientName if ($PatientName);
  
  print "</b></td>
      </tr>
      <tr>
        <td align=center style=\"font-family:arial; font-size: 14px; color: ".$HeadingColour.";\" width=10%>"
      .$StartDate." - ".$EndDate."</td>
      </tr>
  </table>
  <p>
  
  <table bgcolor=".$ReportColour."  border=0 cellspacing=2 cellpadding=0 width=683>
    <tr>
      <td width=50%>
        <table bgcolor=white border=0 cellspacing=0 cellpadding=0 width=100%>
          <tr><td colspan=4><b>".$LocNotes."</b></td></tr>
          <tr>
            <td width=10>&nbsp;1.</td><td>&nbsp;</td>
            <td>".$LocNotes1." <b>".$GlucoseUnits."</b></td><td>&nbsp;</td>
          </tr>
          <tr>
           <td width=10>&nbsp;2.</td><td>&nbsp;</td>
            <td bgcolor=LIGHTGOLDENRODYELLOW>".$LocNotes2."</td><td>&nbsp;</td>
          </tr>
          <tr>
           <td width=10>&nbsp;3.</td><td>&nbsp;</td>
            <td>".$LocNotes3."</td><td>&nbsp;</td>
          </tr>
          <tr>
           <td width=10></td><td>&nbsp;</td>
            <td>&nbsp;&nbsp;&nbsp;&nbsp;<font color=navy><b>".$LocInsulinA."</b></font>, <font color=teal><b>".$LocInsulinB."</b></font>, <font color=maroon><b>".$LocCarbs."</b></font>, <font color=purple><b>".$LocMedicine."</b></font>.</td><td>&nbsp;</td>
          </tr>
          <tr>
           <td width=10>&nbsp;4.</td><td>&nbsp;</td>
           <td>".$LocNotes4."</td><td>&nbsp;</td>
          </tr>
        </table
      </td>
      <td width=50%>
        <table bgcolor=white border=0 cellspacing=0 cellpadding=0 width=100%>
          <tr><td colspan=4><b>".$LocGluc."</b></td></tr>
          <tr>
            <td width=20 bgcolor=".$GlucoseColors{"VeryHigh"}.">&nbsp;</td><td>&nbsp;</td>
            <td>".$LocGluc1a." <b>".$LocGluc1b." ".$GlucoseLevels{"CriticalHigh"}." ".$GlucoseUnits."</b></td><td>&nbsp;</td>
          </tr>
          <tr>
            <td width=20 bgcolor=".$GlucoseColors{"High"}.">&nbsp;</td><td>&nbsp;</td>
            <td>".$LocGluc2a." <b>".$LocGluc2b." ".$GlucoseLevels{"TargetHigh"}." ".$GlucoseUnits."</b></td><td>&nbsp;</td>
          </tr>
          <tr>
            <td width=20 bgcolor=".$GlucoseColors{"Target"}.">&nbsp;</td><td>&nbsp;</td>
            <td>".$LocGluc3a." <b>".$LocGluc3b." ".$GlucoseLevels{"TargetLow"}." ".$LocGluc3c." ".$GlucoseLevels{"TargetHigh"}." ".$GlucoseUnits."</b></td><td>&nbsp;</td>
          </tr>
          <tr>
            <td width=20 bgcolor=".$GlucoseColors{"Low"}.">&nbsp;</td><td>&nbsp;</td>
            <td>".$LocGluc4a." <b>".$LocGluc4b." ".$GlucoseLevels{"TargetLow"}." ".$GlucoseUnits."</b></td><td>&nbsp;</td>
          </tr>
          <tr>
            <td width=20 bgcolor=".$GlucoseColors{"VeryLow"}.">&nbsp;</td><td>&nbsp;</td>
            <td>".$LocGluc5a." <b>".$LocGluc5b." ".$GlucoseLevels{"CriticalLow"}." ".$GlucoseUnits."</b></td><td>&nbsp;</td>
          </tr>
      </table
      </td>
    </tr>
  </table>
  <p>&nbsp;
  ";
  
}


#=======================================
# Report Generation Subroutines - Footer
#=======================================
sub PrintReportFooter() {

  #Close out the HTML
  print "
  <font style=\"font-family: arial; color: grey; font-size: 9px;\">".$LocReportFooter." Diabetes Logbook X (c) Paul Nesfield, 2006 - ".scalar 1900 + (localtime)[5]."</font></center>
  </body>
</html>";

}



#==============================================================================
#
# SUBROUTINES - PRINT REPORTS - GRAPH
#
#==============================================================================

#=======================================
# Report Generation Subroutines - Graph
#=======================================
sub PrintGraph() {

  #Establish variables for storing glucose data
  my $AvgGlucose;
  my $NumEvents;
  
  #Setup graph defaults
  if ($GlucoseUnits eq "mmol/L") {
    $GraphCriticalHigh=sprintf("%0.0f", 110 - scalar $GlucoseLevels{"CriticalHigh"} * 5);
    $GraphTargetHigh=sprintf("%0.0f", 110 - scalar $GlucoseLevels{"TargetHigh"} * 5);
    $GraphTargetLow=sprintf("%0.0f", 110 - scalar $GlucoseLevels{"TargetLow"} * 5);
    $GraphCriticalLow=sprintf("%0.0f", 110 - scalar $GlucoseLevels{"CriticalLow"} * 5);
  }
  else {
    $GraphCriticalHigh=sprintf("%0.0f", 110 - (scalar $GlucoseLevels{"CriticalHigh"} * 0.25));
    $GraphTargetHigh=sprintf("%0.0f", 110 - (scalar $GlucoseLevels{"TargetHigh"} * 0.25));
    $GraphTargetLow=sprintf("%0.0f", 110 - (scalar $GlucoseLevels{"TargetLow"} * 0.25));
    $GraphCriticalLow=sprintf("%0.0f", 110 - (scalar $GlucoseLevels{"CriticalLow"} * 0.25));
  }
  
  #Get reverse order of glucose timestamps
  (@GraphTimestamps)=sort InverseSort (keys %Glucose);
  
  #To strech out the graph scale to fill 300 columns we need multipler
  #$NumEvents is how many table columns per glucose event
  $NumEvents=1;
  if (@GraphTimestamps) {
    $NumEvents=scalar sprintf("%0.0f", 300 / @GraphTimestamps );
    #$NumEvents=0;
  }

  #Build array of scaled glucose data, padded with blank colums if needed
  #Record timestamp for each column
  #Remember that the array is ordered newest events first
  undef @GraphGlucose;
  $i=0;  #Number of columns populated
  $ge=0; #Actual number of events
  $AvgGraphGlucose=0;
  $AvgGraphGlucoseActual=0;
  foreach (@GraphTimestamps) {
    if ($GlucoseUnits eq "mmol/L") {
      if (scalar $Glucose{$_} >= 22.0) {
        $GraphGlucose="1";
      }
      elsif (scalar $Glucose{$_} <= 0.0) {
        $GraphGlucose="110";
      }
      else {
        $GraphGlucose=sprintf("%0.0f", 110 - scalar $Glucose{$_} * 5);
      }
    }
    else {
      if (scalar $Glucose{$_} >= 440) {
        $GraphGlucose="1";
      }
      elsif (scalar $Glucose{$_} <= 0.0) {
        $GraphGlucose="110";
      }
      else {
        $GraphGlucose=sprintf("%0.0f", 110 - scalar $Glucose{$_} * 0.25);
      }
    }
    $ge++;
    $i++;
    $GraphGlucose[$i]=$GraphGlucose;
    $GraphTimestamp[$i]=$_;
    $AvgGraphGlucose += $GraphGlucose;
    $AvgGraphGlucoseActual += $Glucose{$_};
    for ($e = 1; $e < $NumEvents; $e++) {
      $i++;
      $GraphGlucose[$i]="";
      $GraphTimestamp[$i]=$_;
    }
  }
  
  if ($i > 0) {
    if ($GlucoseUnits eq "mmol/L") {
      $EstHbA1c = sprintf("%0.1f", ((($AvgGraphGlucoseActual/$i) * 18) + 77.3)/35.6);
    }
    else {
      $EstHbA1c = sprintf("%0.1f", (($AvgGraphGlucoseActual/$i) + 77.3)/35.6);
    }
  }
  
  #Set max number of glucose evenets
  $i=300 if ($i > 300);
  
  #Set first and last date/times for graph axis
  $TimeMax=$Date{$GraphTimestamp[1]}." ".$Time{$GraphTimestamp[1]};
  $TimeMin=$Date{$GraphTimestamp[$i]}." ".$Time{$GraphTimestamp[$i]};
  
  #Add note that only 300 events are used
  $NumReadingsText="";
  $NumReadingsText="
          <tr>
            <td valign=top align=right><font style=\"font-family:arial; font-size: 10px; color: grey;\" >".$LocGraph5."&nbsp;</font></td>
          </tr>" if ($ge >= 300);
    
  #Calculate the average glucose
  if ($ge >= 1) {
    $AvgGraphGlucose=sprintf("%0.0f", scalar $AvgGraphGlucose / $ge);
    if ($GlucoseUnits eq "mmol/L") {
      $AvgGraphGlucoseActual=sprintf("%0.1f", scalar $AvgGraphGlucoseActual / $ge);
    }
    else {
      $AvgGraphGlucoseActual=sprintf("%0.0f", scalar $AvgGraphGlucoseActual / $ge);
    }
  }
   
  #Now to load a hash table for the columns ($x) using the
  #glucose data. Also add colour for different days, average, and boundaries
  undef @GraphRow;
  for ($x = 1; $x <= $i; $x++) {
  
    #Get day from timestamp
    $GraphEventDate=substr $GraphTimestamp[$x], 0, 8;
    
    #Reset variables for each row
    if ($x eq 1) {
      $LastDate=$GraphEventDate;
      $AddColumnColor="";
    }
    
    #Change column colour if the day changes
    if ($LastDate ne $GraphEventDate) {
      if ($AddColumnColor eq "") {
        $AddColumnColor=" bgcolor=lightcyan ";
      }
      else {
        $AddColumnColor="";
      }
    }
    
    $CritMark="<tr><td height=1 bgcolor=".$GlucoseColors{"VeryHigh"}."></td></tr><tr><td height=1></td></tr>\n";
    $TargMark="<tr><td height=1 bgcolor=".$GlucoseColors{"Target"}."></td></tr><tr><td height=1></td></tr>\n";
    $AvgMark="<tr><td height=1 bgcolor=blueviolet></td></tr><tr><td height=1></td></tr>\n";
    $SpotMark="<tr><td height=2 bgcolor=black width =100%></td></tr>\n";
    $ColTitle="title=\"".$Glucose{$GraphTimestamp[$x]}." \n ".$Date{$GraphTimestamp[$x]}." ".$Time{$GraphTimestamp[$x]}."\"";

    $GraphRow[$x]="\n<td><table ".$AddColumnColor." border=0 cellspacing=0 cellpadding=0 align=center height=220 width=100% ".$ColTitle." >\n";
    undef %GlucValues;
    undef @OrderedGValues;
    $GlucValues{$GraphCriticalLow}="cLOW" if ($GraphCriticalLow);
    $GlucValues{$GraphTargetLow}="tLOW" if ($GraphTargetLow);
    $GlucValues{$GraphTargetHigh}="tHIGH" if ($GraphTargetHigh);
    $GlucValues{$GraphCriticalHigh}="cHIGH" if ($GraphCriticalHigh);    
    $GlucValues{$AvgGraphGlucose}="AVERAGE" if ($AvgGraphGlucose);
    $GlucValues{$GraphGlucose[$x]}="GLUCOSE" if ($GraphGlucose[$x]);
    (@OrderedGValues)= sort NormalSortNum (keys %GlucValues);

    $LastValue=0;
    $CurrentValue=0;
    foreach (@OrderedGValues) {
      $CurrentValue=$_;
      $GraphRow[$x].="<tr><td width=100% height=".((scalar $CurrentValue - $LastValue) * 2)."></td></tr>\n" if ($CurrentValue > $LastValue);
      if ($GlucValues{$CurrentValue} eq "GLUCOSE") {
        $GraphRow[$x].=$SpotMark;
      }
      if ($GlucValues{$CurrentValue} eq "AVERAGE") {
        $GraphRow[$x].=$AvgMark;
      }
      if (($GlucValues{$CurrentValue} eq "cHIGH") || $GlucValues{$CurrentValue} eq "cLOW") {
        $GraphRow[$x].=$CritMark;
      }
      if (($GlucValues{$CurrentValue} eq "tHIGH") || $GlucValues{$CurrentValue} eq "tLOW") {
        $GraphRow[$x].=$TargMark;
      }
      $LastValue=$CurrentValue + 1;
    }
    $GraphRow[$x].="<tr><td width=100% height=".((scalar 110 - $LastValue) * 2)."></td></tr>\n";

    $GraphRow[$x].="</table></td>";
        
    #Save this date for next go round
    $LastDate=$GraphEventDate;
  }

  #print "<br>Estimated HbA1c = ".$EstHbA1c."%<br>";
  #Print the report header
  print "
  <a name=GRAPH></a>
  <table bgcolor=".$ReportColour."  border=0 cellspacing=0 cellpadding=2 width=683>
    <tr>
      <td>
        <table  bgcolor=white border=0 cellspacing=0 cellpadding=0 width=100%>
          <tr>
            <th bgcolor=".$ReportColour.">".$LocGraphHeader."</th>
          </tr>
          <tr>
            <td height=12>
            </td>
          </tr>
          <tr>
            <td>
              <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=640 height=240>
                <tr>
                  <td width=34 height=221 bgcolor=white rowspan=3>
                  ".$GraphYscale."
                  </td>
                  <td width=6 height=1 bgcolor=grey colspan=2></td>
                  <td width=600 height=220 bgcolor=white colspan=2 rowspan=2>
                    <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=600 height=220>";

  #Print each log entry from the Hash tables, extract each one by the timestamp
  
  #Print row by row
  $blankspace= 300 - scalar $i;
    
  #Start new row
  print "
                      <tr width=600 height=220 valign=bottom>";
                    
  #Print oldest to newest (left to right) so it aligns with the LH axis
  for ($x = $i; $x >= 1; $x--) {
    print $GraphRow[$x];
  }
  
  print "<td width=".$blankspace."></td>";
  #End row
  print "</tr>";

  #End the table
  print "
                    </table>
                  </td>
                </tr>
                <tr>
          <td width=5 height=219 bgcolor=white>
            <table bgcolor=white border=0 cellspacing=0 cellpadding=0 height=219>
              <tr><td colspan=2 width=3 height=19></td></tr>
              <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=19></td></tr>
              <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=19></td></tr>
              <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=19></td></tr>
              <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=19></td></tr>
              <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=19></td></tr>
              <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=19></td></tr>
              <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=19></td></tr>
              <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=19></td></tr>
              <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=19></td></tr>
              <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=19></td></tr>
            </table>
          </td>
          <td width=1 height=219 bgcolor=grey></td>
                </tr>
                <tr>
                  <td width=606 height=1 bgcolor=grey colspan=4></td>
                </tr>
                <tr>
                  <td width=34 height=5 bgcolor=white></td>
                  <td width=5 height=5 bgcolor=white></td>
                  <td width=1 height=5 bgcolor=grey></td>
                  <td width=599 height=5 bgcolor=white></td>
                  <td width=1 height=5 bgcolor=grey></td>
                </tr>
                <tr>
                  <td width=39 height=14 bgcolor=white colspan=2></td>
                  <td width=601 height=14 bgcolor=white colspan=3>
                    <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=100% height=100%>
                      <tr>
                        <td valign=top align=left width=50%>".$TimeMin."</td>
                        <td valign=top align=right width=50%>".$TimeMax."</td>
                      </tr>
                    </table>
                  </td>
                </tr>
              </table>
            </td>
          </tr>".$NumReadingsText."
          <tr>
            <td>
              <table bgcolor=white border=0 cellspacing=0 cellpadding=0>
                <tr><td colspan=4><b>".$LocGraph0."</b></td></tr>
                <tr>
                  <td width=5></td>
                  <td valign=middle width=10>
                    <table bgcolor=white border=0 cellspacing=0 cellpadding=0>
                      <tr><td width=10 height=2 bgcolor=".$GlucoseColors{"VeryHigh"}." width=100%></td></tr></table>
                  </td>
                  <td>".$LocGraph1.",</td>
                  <td width=5></td>
                  <td valign=middle width=10>
                    <table bgcolor=white border=0 cellspacing=0 cellpadding=0>
                      <tr><td width=10 height=2 bgcolor=".$GlucoseColors{"Target"}." width=100%></td></tr></table>
                  </td>
                  <td>".$LocGraph2.",</td>
                  <td width=5></td>
                  <td valign=middle width=10>
                    <table bgcolor=white border=0 cellspacing=0 cellpadding=0>
                      <tr><td width=10 height=2 bgcolor=blueviolet width=100%></td></tr></table>
                  </td>
                  <td> [".$AvgGraphGlucoseActual."] ".$LocGraph3.",</td>
                  <td width=5></td>
                  <td width=10 bgcolor=lightcyan></td>
                  <td>".$LocGraph4."</td><td width=5></td>
                </tr>
              </table
            </td>
          </tr>
        </table>
      </td>
    </tr>
  </table>
  <p>&nbsp;
  ";

}


#=======================================
# Report Generation Subroutines - Graph
#=======================================
sub PrintGraphHbA1c() {

  #Establish variables for storing glucose data
  my $NumEvents;
  my $HbA1cScheme;
  
  #Retrieve the HbA1c scheme direct from the Mac OS X defaults system
  open(HBA1C, "/usr/bin/defaults read com.Nesfield.DiabetesLogbookX HbA1cStandard|");
  $HbA1cScheme = <HBA1C>;
  close(HBA1C);
  
  chomp $HbA1cScheme;

  #Get maximum HbA1c value
  $MaxHbA1c=(sort InverseSortNum (values %HbA1c))[0];

  #If the max HbA1c is below this we will assume old % values
  #If above we'll use mmol/mol and convert any value under this to the new units.
  $HbA1cUnitThreshold=15.0;
  
  #Setup graph defaults
  if ($MaxHbA1c >= $HbA1cUnitThreshold) {
    $ReCalcHbA1c="YES";
  	$HbA1cUnits="mmol<br/>/mol";
  	$HbA1cUnits2="mmol/mol";
	  $HbA1cFactorUp=10;
		$HbA1cLow=30;
		$HbA1cHigh=85;
		$HbA1cTarget=48;
    $HbA1cFactorUpGraph=2;
		if ($HbA1cScheme eq "JDS/JSCC") {
		  $HbA1cDivisor=0.927;
		  $HbA1cDelta=1.73;
		}
		elsif ($HbA1cScheme eq "Swedish") {
		  $HbA1cDivisor=0.989;
		  $HbA1cDelta=0.88;
		}
		else {
		  $HbA1cDivisor=0.915;
		  $HbA1cDelta=2.15;
		}
	}
  else {
    $ReCalcHbA1c="NO";
  	$HbA1cUnits="%";
  	$HbA1cUnits2="%";
		$HbA1cDivisor=1.0;
		$HbA1cDelta=0.0;
		$HbA1cFactorUp=1.0;
    $HbA1cFactorUpGraph=20;
    if ($HbA1cScheme eq "JDS/JSCC") {
		  $HbA1cLow=4.5;
		  $HbA1cHigh=10.0;
		  $HbA1cTarget=6.2;
		}
		elsif ($HbA1cScheme eq "Swedish") {
		  $HbA1cLow=4.0;
		  $HbA1cHigh=9.5;
		  $HbA1cTarget=5.6;
		}
		else {
		  $HbA1cLow=5.0;
		  $HbA1cHigh=10.5;
		  $HbA1cTarget=6.5;
		}
  }
  
  $GraphMax="< ".$HbA1cHigh." ".$HbA1cUnits;
  $GraphMin="> ".$HbA1cLow;
  $GraphTarget=scalar sprintf("%0.0f", ($HbA1cHigh - $HbA1cTarget)*110/($HbA1cHigh - $HbA1cLow) ); #=48mmol/mol
  
  #Get reverse order of timestamps
  (@GraphTimestamps)=sort InverseSort (keys %HbA1c);
  
  #To strech out the graph scale to fill 150 columns we need multipler
  #$NumEvents is how many table columns per glucose event
  $NumEvents=1;
  if (@GraphTimestamps) {
    $NumEvents=scalar sprintf("%0.0f", 150 / @GraphTimestamps );
  }

  #Build array of scaled data, padded with blank colums if needed
  #Record timestamp for each column
  #Remember that the array is ordered newest events first
  undef @GraphEvents;
  undef %HbA1cSpotColour;
  undef %HbA1cOrig;
  $i=0;  #Number of columns populated
  $ge=0; #Actual number of events
  foreach (@GraphTimestamps) {
 
    #Manipulate the data to the right units. Save a note  of the old value for the graph title
    $HbA1cOrig{$_}=" [ = ".$HbA1c{$_}."% (".$HbA1cScheme.") ]";
    if (scalar $HbA1c{$_} <= $HbA1cUnitThreshold && ($ReCalcHbA1c eq "YES")) {
      #Need to recalculate the HbA1c into mmol/mol
      $HbA1c{$_}=sprintf("%0.0f", $HbA1cFactorUp * ((scalar $HbA1c{$_} - $HbA1cDelta) / $HbA1cDivisor));
      $HbA1cSpotColour{$_}="purple";
    }
    else {
      $HbA1c{$_}=sprintf("%0.0f", scalar $HbA1c{$_});
      $HbA1cSpotColour{$_}="black";
      $HbA1cOrig{$_}="";
    }
    
    if (scalar $HbA1c{$_} >= $HbA1cHigh) {
      $GraphEvent="1";
    }
    elsif (scalar $HbA1c{$_} <= $HbA1cLow) {
      $GraphEvent="110";
    }
    else {
      $GraphEvent=sprintf("%0.0f", 110 - scalar ($HbA1c{$_} - $HbA1cLow) * $HbA1cFactorUpGraph);
    }
    $ge++;
    $i++;
    $GraphEvent[$i]=$GraphEvent;
    $GraphTimestamp[$i]=$_;
    for ($e = 1; $e < $NumEvents; $e++) {
      $i++;
      $GraphEvent[$i]="";
      $GraphTimestamp[$i]=$_;
    }
  }
  
  if ($ge > 0) {
    #Set max number of glucose evenets
    $i=150 if ($i > 150);
    
    #Set first and last date/times for graph axis
    $TimeMax=$Date{$GraphTimestamp[1]}." ".$Time{$GraphTimestamp[1]};
    $TimeMin=$Date{$GraphTimestamp[$i]}." ".$Time{$GraphTimestamp[$i]};
    
    #Add note that only 300 events are used
    $NumReadingsText="";
    $NumReadingsText="
        <tr>
        <td valign=top align=right><font style=\"font-family:arial; font-size: 10px; color: grey;\" >".$LocGraph7."&nbsp;</font></td>
        </tr>" if ($ge >= 150);

    #Now to load a hash table for the columns ($x) using the
    # data. Also add colour for different days, average, and boundaries
    undef %GraphRow;
    for ($x = 1; $x <= $i; $x++) {
  
      #Get day from timestamp
      $GraphEventDate=substr $GraphTimestamp[$x], 0, 8;
    
      #Reset variables for each row
      if ($x eq 1) {
        $LastDate=$GraphEventDate;
        $AddColumnColor="";
      }
    
      #Change column colour if the day changes
      if ($LastDate ne $GraphEventDate) {
        if ($AddColumnColor eq "") {
          $AddColumnColor=" bgcolor=lightcyan";
        }
        else {
          $AddColumnColor="";
        }
      }
    
      $TargMark="<tr><td height=1 bgcolor=".$GlucoseColors{"Target"}."></td></tr><tr><td height=1></td></tr>\n";
      $SpotMark="<tr><td height=2 bgcolor=".$HbA1cSpotColour{$GraphTimestamp[$x]}." width =100%></td></tr>\n";
      $ColTitle="title=\"".$HbA1c{$GraphTimestamp[$x]}.$HbA1cOrig{$GraphTimestamp[$x]}."\n".$Date{$GraphTimestamp[$x]}." ".$Time{$GraphTimestamp[$x]}."\"";

      $GraphRow[$x]="\n<td><table ".$AddColumnColor." border=0 cellspacing=0 cellpadding=0 align=center height=220 width=100% ".$ColTitle." >\n";
      undef %HbValues;
      undef @OrderedHbValues;
      $HbValues{$GraphTarget}="TARGET" if ($GraphTargetHigh);
      $HbValues{$GraphEvent[$x]}="HBA1C" if ($GraphEvent[$x]);
      (@OrderedGValues)= sort NormalSortNum (keys %HbValues);

      $LastValue=1;
      $CurrentValue=1;
      foreach (@OrderedGValues) {
        $CurrentValue=$_;
        $GraphRow[$x].="<tr><td width=100% height=".((scalar $CurrentValue - $LastValue) * 2)."></td></tr>\n" if ($CurrentValue > $LastValue);
        if ($HbValues{$CurrentValue} eq "HBA1C") {
          $GraphRow[$x].=$SpotMark;
        }
        if ($HbValues{$CurrentValue} eq "TARGET") {
          $GraphRow[$x].=$TargMark;
        }
        $LastValue=$CurrentValue + 1;
      }
      $GraphRow[$x].="<tr><td width=100% height=".((scalar 110 - $LastValue) * 2)."></td></tr>\n";

      $GraphRow[$x].="</table></td>";
    
      #Save this date for next go round
      $LastDate=$GraphEventDate;
    }

    #Print the report header
    print "
    <a name=GRAPHHBA1C></a>
    <table bgcolor=".$ReportColour."  border=0 cellspacing=0 cellpadding=2 width=683>
    <tr>
      <td>
      <table  bgcolor=white border=0 cellspacing=0 cellpadding=0 width=100%>
        <tr>
        <th bgcolor=".$ReportColour.">".$LocGraphHbA1cHeader."</th>
        </tr>
        <tr>
        <td height=12>
        </td>
        </tr>
        <tr>
        <td>
          <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=640 height=240>
          <tr>
            <td width=34 height=221 bgcolor=white rowspan=3>
            <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=100% height=100%>
              <tr>
              <td valign=top align=right height=50%>".$GraphMax."</td>
              </tr>
              <tr>
              <td valign=bottom align=right height=50%>".$GraphMin."</td>
              </tr>
            </table>
            </td>
            <td width=6 height=1 bgcolor=grey colspan=2></td>
            <td width=600 height=220 bgcolor=white colspan=2 rowspan=2>
            <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=600 height=220>";

            #Print each log entry from the Hash tables, extract each one by the timestamp

            #Print row by row
            $blankspace= 150 - scalar $i;

            #Start new row
            print "
                                <tr width=600 height=220 valign=bottom>";

            #Print oldest to newest (left to right) so it aligns with the LH axis
            for ($x = $i; $x >= 1; $x--) {
              print $GraphRow[$x];
            }

            print "<td width=".$blankspace."></td>";
            #End row
            print "</tr>";
            
            #End the table
            print "
            </table>
            </td>
          </tr>
          <tr>
            <td width=5 height=219 bgcolor=white>
              <table bgcolor=white border=0 cellspacing=0 cellpadding=0 height=219>
                <tr><td colspan=2 width=3 height=19></td></tr>
                <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=19></td></tr>
                <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=19></td></tr>
                <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=19></td></tr>
                <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=19></td></tr>
                <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=19></td></tr>
                <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=19></td></tr>
                <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=19></td></tr>
                <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=19></td></tr>
                <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=19></td></tr>
                <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=19></td></tr>
              </table>
            </td>
            <td width=1 height=219 bgcolor=grey></td>
          </tr>
          <tr>
            <td width=606 height=1 bgcolor=grey colspan=4></td>
          </tr>
          <tr>
            <td width=34 height=5 bgcolor=white></td>
            <td width=5 height=5 bgcolor=white></td>
            <td width=1 height=5 bgcolor=grey></td>
            <td width=599 height=5 bgcolor=white></td>
            <td width=1 height=5 bgcolor=grey></td>
          </tr>
          <tr>
            <td width=39 height=14 bgcolor=white colspan=2></td>
            <td width=601 height=14 bgcolor=white colspan=3>
            <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=100% height=100%>
              <tr>
              <td valign=top align=left width=50%>".$TimeMin."</td>
              <td valign=top align=right width=50%>".$TimeMax."</td>
              </tr>
            </table>
            </td>
          </tr>
          </table>
        </td>
        </tr>".$NumReadingsText."
        <tr>
        <td>
          <table bgcolor=white border=0 cellspacing=0 cellpadding=0>
          <tr><td colspan=4><b>".$LocGraph0."</b></td></tr>
          <tr>
            <td width=5></td>
            <td valign=middle width=10>
            <table bgcolor=white border=0 cellspacing=0 cellpadding=0>
              <tr><td width=10 height=2 bgcolor=".$GlucoseColors{"Target"}." width=100%></td></tr></table>
            </td>
            <td>[".$HbA1cTarget."] ".$LocGraph6.",</td>
            <td width=5></td>
            <td width=10 bgcolor=lightcyan></td>
            <td>".$LocGraph4."</td><td width=5></td>
          </tr>
          </table
        </td>
        </tr>
      </table>
      </td>
    </tr>
    </table>
    <p>&nbsp;
    ";
  }
}


#=======================================
# Report Generation Subroutines - Graph
#=======================================
sub PrintGraphWeight() {

  #Establish variables for storing glucose data
  my $NumEvents;
    
  #Get reverse order of timestamps
  (@GraphTimestamps)=sort InverseSort (keys %Weight);
  
  #To strech out the graph scale to fill 150 columns we need multipler
  #$NumEvents is how many table columns per glucose event
  $NumEvents=1;
  if (@GraphTimestamps) {
    $NumEvents=scalar sprintf("%0.0f", 150 / @GraphTimestamps );
  }

  #Build array of scaled data, padded with blank colums if needed
  #Record timestamp for each column
  #Remember that the array is ordered newest events first
  undef @GraphEvents;
  $i=0;  #Number of columns populated
  $ge=0; #Actual number of events
  undef $GraphMax;
  undef $GraphMin;
  foreach (@GraphTimestamps) {
    $GraphMax=$Weight{$_} if (!$GraphMax || $GraphMax < $Weight{$_});
    $GraphMin=$Weight{$_} if (!$GraphMin || $GraphMin > $Weight{$_});
  }
  $GraphMin=$GraphMax - 1 if ($GraphMax == $GraphMin);
  if ($GraphMax) {
    foreach (@GraphTimestamps) {
      $GraphEvent=sprintf("%0.0f", 110 - scalar (109*($Weight{$_} - $GraphMin)/($GraphMax - $GraphMin)));
      $ge++;
      $i++;
      $GraphEvent[$i]=$GraphEvent;
      $GraphTimestamp[$i]=$_;
      for ($e = 1; $e < $NumEvents; $e++) {
        $i++;
        $GraphEvent[$i]="";
        $GraphTimestamp[$i]=$_;
      }
    }
  }
  
  if ($ge > 0) {
    #Set max number of glucose evenets
    $i=150 if ($i > 150);
    
    #Set first and last date/times for graph axis
    $TimeMax=$Date{$GraphTimestamp[1]}." ".$Time{$GraphTimestamp[1]};
    $TimeMin=$Date{$GraphTimestamp[$i]}." ".$Time{$GraphTimestamp[$i]};
    
    #Add note that only 300 events are used
    $NumReadingsText="";
    $NumReadingsText="
        <tr>
        <td valign=top align=right><font style=\"font-family:arial; font-size: 10px; color: grey;\" >".$LocGraph7."&nbsp;</font></td>
        </tr>" if ($ge >= 150);

    #Now to load a hash table for the columns ($x) using the
    # data. Also add colour for different days, average, and boundaries
    undef %GraphRow;
    for ($x = 1; $x <= $i; $x++) {
  
      #Get day from timestamp
      $GraphEventDate=substr $GraphTimestamp[$x], 0, 8;
    
      #Reset variables for each row
      if ($x eq 1) {
        $LastDate=$GraphEventDate;
        $AddColumnColor="";
      }
    
      #Change column colour if the day changes
      if ($LastDate ne $GraphEventDate) {
        if ($AddColumnColor eq "") {
          $AddColumnColor=" bgcolor=lightcyan";
        }
        else {
          $AddColumnColor="";
        }
      }
    
      $SpotMark="<tr><td height=2 bgcolor=black width =100%></td></tr>\n";
      $ColTitle="title=\"".$Weight{$GraphTimestamp[$x]}." \n ".$Date{$GraphTimestamp[$x]}." ".$Time{$GraphTimestamp[$x]}."\"";

      $GraphRow[$x]="\n<td><table ".$AddColumnColor." border=0 cellspacing=0 cellpadding=0 align=center height=220 width=100% ".$ColTitle." >\n";
      undef %WtValues;
      undef @OrderedWtValues;
      $WtValues{$GraphEvent[$x]}="WEIGHT" if ($GraphEvent[$x]);
      (@OrderedWtValues)= sort NormalSortNum (keys %WtValues);

      $LastValue=1;
      $CurrentValue=1;
      foreach (@OrderedWtValues) {
        $CurrentValue=$_;
        $GraphRow[$x].="<tr><td width=100% height=".((scalar $CurrentValue - $LastValue) * 2)."></td></tr>\n" if ($CurrentValue > $LastValue);
        if ($WtValues{$CurrentValue} eq "WEIGHT") {
          $GraphRow[$x].=$SpotMark;
        }
        $LastValue=$CurrentValue + 1;
      }
      $GraphRow[$x].="<tr><td width=100% height=".((scalar 110 - $LastValue) * 2)."></td></tr>\n";

      $GraphRow[$x].="</table></td>";
      
      #Save this date for next go round
      $LastDate=$GraphEventDate;
    }

    #Print the report header
    print "
    <a name=GRAPHWEIGHT></a>
    <table bgcolor=".$ReportColour."  border=0 cellspacing=0 cellpadding=2 width=683>
    <tr>
      <td>
      <table  bgcolor=white border=0 cellspacing=0 cellpadding=0 width=100%>
        <tr>
        <th bgcolor=".$ReportColour.">".$LocGraphWeightHeader."</th>
        </tr>
        <tr>
        <td height=12>
        </td>
        </tr>
        <tr>
        <td>
          <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=640 height=240>
          <tr>
            <td width=34 height=221 bgcolor=white rowspan=3>
            <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=100% height=100%>
              <tr>
              <td valign=top align=right height=50%>".$GraphMax." ".$WeightUnits."</td>
              </tr>
              <tr>
              <td valign=bottom align=right height=50%>".$GraphMin."</td>
              </tr>
            </table>
            </td>
            <td width=6 height=1 bgcolor=grey colspan=2></td>
            <td width=600 height=220 bgcolor=white colspan=2 rowspan=2>
            <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=600 height=220>";

            #Print each log entry from the Hash tables, extract each one by the timestamp

            #Print row by row
            $blankspace= 150 - scalar $i;

            #Start new row
            print "
                                <tr width=600 height=220 valign=bottom>";

            #Print oldest to newest (left to right) so it aligns with the LH axis
            for ($x = $i; $x >= 1; $x--) {
              print $GraphRow[$x];
            }

            print "<td width=".$blankspace."></td>";
            #End row
            print "</tr>";
            
            #End the table
            print "
            </table>
            </td>
          </tr>
          <tr>
            <td width=5 height=219 bgcolor=white>
              <table bgcolor=white border=0 cellspacing=0 cellpadding=0 height=219>
                <tr><td colspan=2 width=5 height=21></td></tr>
                <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=21></td></tr>
                <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=21></td></tr>
                <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=21></td></tr>
                <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=21></td></tr>
                <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=21></td></tr>
                <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=21></td></tr>
                <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=21></td></tr>
                <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=21></td></tr>
                <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=21></td></tr>
              </table>
            </td>
            <td width=1 height=219 bgcolor=grey></td>
          </tr>
          <tr>
            <td width=606 height=1 bgcolor=grey colspan=4></td>
          </tr>
          <tr>
            <td width=34 height=5 bgcolor=white></td>
            <td width=5 height=5 bgcolor=white></td>
            <td width=1 height=5 bgcolor=grey></td>
            <td width=599 height=5 bgcolor=white></td>
            <td width=1 height=5 bgcolor=grey></td>
          </tr>
          <tr>
            <td width=39 height=14 bgcolor=white colspan=2></td>
            <td width=601 height=14 bgcolor=white colspan=3>
            <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=100% height=100%>
              <tr>
              <td valign=top align=left width=50%>".$TimeMin."</td>
              <td valign=top align=right width=50%>".$TimeMax."</td>
              </tr>
            </table>
            </td>
          </tr>
          </table>
        </td>
        </tr>".$NumReadingsText."
        <tr>
        <td>
          <table bgcolor=white border=0 cellspacing=0 cellpadding=0>
          <tr><td colspan=4><b>".$LocGraph0."</b></td></tr>
          <tr>
            <td width=5></td>
            <td width=10 bgcolor=lightcyan></td>
            <td>".$LocGraph4."</td><td width=5></td>
          </tr>
          </table
        </td>
        </tr>
      </table>
      </td>
    </tr>
    </table>
    <p>&nbsp;
    ";
  }
}


#=======================================
# Report Generation Subroutines - Graph
#=======================================
sub PrintGraphHeight() {

  #Establish variables for storing glucose data
  my $NumEvents;
  
  #Get reverse order of timestamps
  (@GraphTimestamps)=sort InverseSort (keys %Height);
  
  #To strech out the graph scale to fill 150 columns we need multipler
  #$NumEvents is how many table columns per glucose event
  $NumEvents=1;
  $ShowEvents=150;
  if (@GraphTimestamps) {
    $NumEvents=scalar sprintf("%0.0f", $ShowEvents / @GraphTimestamps );
    #$NumEvents=@GraphTimestamps + 1;
  }

  #Build array of scaled data, padded with blank colums if needed
  #Record timestamp for each column
  #Remember that the array is ordered newest events first
  undef @GraphEvents;
  $i=0;  #Number of columns populated
  $ge=0; #Actual number of events
  undef $GraphMax;
  undef $GraphMin;
  foreach (@GraphTimestamps) {
    $GraphMax=$Height{$_} if (!$GraphMax || $GraphMax < $Height{$_});
    $GraphMin=$Height{$_} if (!$GraphMin || $GraphMin > $Height{$_});
  }
  $GraphMin=$GraphMax - 1 if ($GraphMax == $GraphMin);
  if ($GraphMax) {
  foreach (@GraphTimestamps) {
      $GraphEvent=sprintf("%0.0f", 110 - scalar (109*($Height{$_} - $GraphMin)/($GraphMax - $GraphMin)));
      $ge++;
      $i++;
      $GraphEvent[$i]=$GraphEvent;
      $GraphTimestamp[$i]=$_;
      for ($e = 1; $e < $NumEvents; $e++) {
        $i++;
        $GraphEvent[$i]="";
        $GraphTimestamp[$i]=$_;
      }
    }
  }
  
  if ($ge > 0) {
    #Set max number of glucose evenets
    $i=$ShowEvents if ($i > $ShowEvents);
    
    #Set first and last date/times for graph axis
    $TimeMax=$Date{$GraphTimestamp[1]}." ".$Time{$GraphTimestamp[1]};
    $TimeMin=$Date{$GraphTimestamp[$i]}." ".$Time{$GraphTimestamp[$i]};
    
    #Add note that only $ShowEvents events are used
    $NumReadingsText="";
    $NumReadingsText="
        <tr>
        <td valign=top align=right><font style=\"font-family:arial; font-size: 10px; color: grey;\" >".$LocGraph7."&nbsp;</font></td>
        </tr>" if ($ge >= $MaxEventsShow);

    #Now to load a hash table for the columns ($x) using the
    # data. Also add colour for different days, average, and boundaries
    undef %GraphRow;
    for ($x = 1; $x <= $i; $x++) {

      #Get day from timestamp
      $GraphEventDate=substr $GraphTimestamp[$x], 0, 8;

      #Reset variables for each row
      if ($x eq 1) {
        $LastDate=$GraphEventDate;
        $AddColumnColor="";
      }

      #Change column colour if the day changes
      if ($LastDate ne $GraphEventDate) {
        if ($AddColumnColor eq "") {
          $AddColumnColor=" bgcolor=lightcyan";
        }
        else {
          $AddColumnColor="";
        }
      }

      $SpotMark="<tr><td height=2 bgcolor=black width =100%></td></tr>\n";
      $ColTitle="title=\"".$Height{$GraphTimestamp[$x]}." \n ".$Date{$GraphTimestamp[$x]}." ".$Time{$GraphTimestamp[$x]}."\"";

      $GraphRow[$x]="\n<td><table ".$AddColumnColor." border=0 cellspacing=0 cellpadding=0 align=center height=220 width=100% ".$ColTitle." >\n";
      undef %HtValues;
      undef @OrderedHtValues;
      $HtValues{$GraphEvent[$x]}="HEIGHT" if ($GraphEvent[$x]);
      (@OrderedHtValues)= sort NormalSortNum (keys %HtValues);

      $LastValue=1;
      $CurrentValue=1;
      $SpotColor="";
      foreach (@OrderedHtValues) {
        $CurrentValue=$_;
        $GraphRow[$x].="<tr><td width=100% height=".((scalar $CurrentValue - $LastValue) * 2)."></td></tr>\n" if ($CurrentValue > $LastValue);
        if ($HtValues{$CurrentValue} eq "HEIGHT") {
          $GraphRow[$x].=$SpotMark;
          #$SpotColor=" bgcolor=black ";
          $SpotColor="";
        }
        $LastValue=$CurrentValue + 1;
      }
      $GraphRow[$x].="<tr><td width=100% height=".((scalar 110 - $LastValue) * 2)."".$SpotColor."></td></tr>\n";

      $GraphRow[$x].="</table></td>";

      #Save this date for next go round
      $LastDate=$GraphEventDate;
    }

    #Print the report header
    print "
    <a name=GRAPHHEIGHT></a>
    <table bgcolor=".$ReportColour."  border=0 cellspacing=0 cellpadding=2 width=683>
    <tr>
      <td>
      <table  bgcolor=white border=0 cellspacing=0 cellpadding=0 width=100%>
        <tr>
        <th bgcolor=".$ReportColour.">".$LocGraphHeightHeader."</th>
        </tr>
        <tr>
        <td height=12>
        </td>
        </tr>
        <tr>
        <td>
          <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=640 height=240>
          <tr>
            <td width=34 height=221 bgcolor=white rowspan=3>
            <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=100% height=100%>
              <tr>
              <td valign=top align=right height=50%>".$GraphMax." ".$HeightUnits."</td>
              </tr>
                          <tr>
              <td valign=bottom align=right height=50%>".$GraphMin."</td>
              </tr>
            </table>
            </td>
            <td width=6 height=1 bgcolor=grey colspan=2></td>
            <td width=600 height=220 bgcolor=white colspan=2 rowspan=2>
            <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=600 height=220>";

            #Print each log entry from the Hash tables, extract each one by the timestamp

            #Print row by row
            $blankspace= $ShowEvents - scalar $i;

            #Start new row
            print "
                                <tr width=600 height=220 valign=bottom>";

            #Print oldest to newest (left to right) so it aligns with the LH axis
            for ($x = $i; $x >= 1; $x--) {
              print $GraphRow[$x];
            }

            print "<td width=".$blankspace."></td>";
            #End row
            print "</tr>";
            
            #End the table
            print "
            </table>
            </td>
          </tr>
          <tr>
            <td width=5 height=219 bgcolor=white>
            <table bgcolor=white border=0 cellspacing=0 cellpadding=0 height=219>
              <tr><td colspan=2 width=5 height=21></td></tr>
              <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=21></td></tr>
              <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=21></td></tr>
              <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=21></td></tr>
              <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=21></td></tr>
              <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=21></td></tr>
              <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=21></td></tr>
              <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=21></td></tr>
              <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=21></td></tr>
              <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=21></td></tr>
            </table>
            </td>
            <td width=1 height=219 bgcolor=grey></td>
          </tr>
          <tr>
            <td width=606 height=1 bgcolor=grey colspan=4></td>
          </tr>
          <tr>
            <td width=34 height=5 bgcolor=white></td>
            <td width=5 height=5 bgcolor=white></td>
            <td width=1 height=5 bgcolor=grey></td>
            <td width=599 height=5 bgcolor=white></td>
            <td width=1 height=5 bgcolor=grey></td>
          </tr>
          <tr>
            <td width=39 height=14 bgcolor=white colspan=2></td>
            <td width=601 height=14 bgcolor=white colspan=3>
            <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=100% height=100%>
              <tr>
              <td valign=top align=left width=50%>".$TimeMin."</td>
              <td valign=top align=right width=50%>".$TimeMax."</td>
              </tr>
            </table>
            </td>
          </tr>
          </table>
        </td>
        </tr>".$NumReadingsText."
        <tr>
        <td>
          <table bgcolor=white border=0 cellspacing=0 cellpadding=0>
          <tr><td colspan=4><b>".$LocGraph0."</b></td></tr>
          <tr>
            <td width=5></td>
            <td width=10 bgcolor=lightcyan></td>
            <td>".$LocGraph4."</td><td width=5></td>
          </tr>
          </table
        </td>
        </tr>
      </table>
      </td>
    </tr>
    </table>
    <p>&nbsp;
    ";
  }
}


#=======================================
# Report Generation Subroutines - Graph
#=======================================
sub PrintGraphBMI() {

  #Establish variables for storing glucose data
  my $NumEvents;
    
  #Get reverse order of timestamps
  (@GraphTimestamps)=sort InverseSort (keys %BMI);
  
  #To strech out the graph scale to fill 150 columns we need multipler
  #$NumEvents is how many table columns per glucose event
  $NumEvents=1;
  if (@GraphTimestamps) {
    $NumEvents=scalar sprintf("%0.0f", 150 / @GraphTimestamps );
  }

  #Build array of scaled data, padded with blank colums if needed
  #Record timestamp for each column
  #Remember that the array is ordered newest events first
  undef @GraphEvents;
  $i=0;  #Number of columns populated
  $ge=0; #Actual number of events
  undef $GraphMax;
  undef $GraphMin;
  foreach (@GraphTimestamps) {
    $GraphMax=$BMI{$_} if (!$GraphMax || $GraphMax < $BMI{$_});
    $GraphMin=$BMI{$_} if (!$GraphMin || $GraphMin > $BMI{$_});
  }
  $GraphMin=$GraphMax - 1 if ($GraphMax == $GraphMin);
  if ($GraphMax) {
    foreach (@GraphTimestamps) {
      $GraphEvent=sprintf("%0.0f", 110 - scalar (109*($BMI{$_} - $GraphMin)/($GraphMax - $GraphMin)));
      $ge++;
      $i++;
      $GraphEvent[$i]=$GraphEvent;
      $GraphTimestamp[$i]=$_;
      for ($e = 1; $e < $NumEvents; $e++) {
        $i++;
        $GraphEvent[$i]="";
        $GraphTimestamp[$i]=$_;
      }
    }
  }
  
  if ($ge > 0) {
    #Set max number of glucose evenets
    $i=150 if ($i > 150);
    
    #Set first and last date/times for graph axis
    $TimeMax=$Date{$GraphTimestamp[1]}." ".$Time{$GraphTimestamp[1]};
    $TimeMin=$Date{$GraphTimestamp[$i]}." ".$Time{$GraphTimestamp[$i]};
    
    #Add note that only 300 events are used
    $NumReadingsText="";
    $NumReadingsText="
        <tr>
        <td valign=top align=right><font style=\"font-family:arial; font-size: 10px; color: grey;\" >".$LocGraph7."&nbsp;</font></td>
        </tr>" if ($ge >= 150);

    #Now to load a hash table for the columns ($x) using the
    # data. Also add colour for different days, average, and boundaries
    undef %GraphRow;
    for ($x = 1; $x <= $i; $x++) {
  
      #Get day from timestamp
      $GraphEventDate=substr $GraphTimestamp[$x], 0, 8;
    
      #Reset variables for each row
      if ($x eq 1) {
        $LastDate=$GraphEventDate;
        $AddColumnColor="";
      }
    
      #Change column colour if the day changes
      if ($LastDate ne $GraphEventDate) {
        if ($AddColumnColor eq "") {
          $AddColumnColor=" bgcolor=lightcyan";
        }
        else {
          $AddColumnColor="";
        }
      }
    
      $SpotMark="<tr><td height=2 bgcolor=black width =100%></td></tr>\n";
      if ($Height{$GraphTimestamp[$x]}) {
        $ColTitle="title=\"".$BMI{$GraphTimestamp[$x]}." \n ".$LocWeight."=".$Weight{$GraphTimestamp[$x]}.$WeightUnits." \n ".$LocHeight."=".$Height{$GraphTimestamp[$x]}.$HeightUnits." \n ".$Date{$GraphTimestamp[$x]}." ".$Time{$GraphTimestamp[$x]}."\"";
      }
      else {
        $ColTitle="title=\"".$BMI{$GraphTimestamp[$x]}." \n ".$LocWeight."=".$Weight{$GraphTimestamp[$x]}.$WeightUnits." \n <- ".$LocHeight."\n ".$Date{$GraphTimestamp[$x]}." ".$Time{$GraphTimestamp[$x]}."\"";
      }

      $GraphRow[$x]="\n<td><table ".$AddColumnColor." border=0 cellspacing=0 cellpadding=0 align=center height=220 width=100% ".$ColTitle." >\n";
      undef %BMIValues;
      undef @OrderedBMIValues;
      $BMIValues{$GraphEvent[$x]}="BMI" if ($GraphEvent[$x]);
      (@OrderedBMIValues)= sort NormalSortNum (keys %BMIValues);

      $LastValue=1;
      $CurrentValue=1;
      foreach (@OrderedBMIValues) {
        $CurrentValue=$_;
        $GraphRow[$x].="<tr><td width=100% height=".((scalar $CurrentValue - $LastValue) * 2)."></td></tr>\n" if ($CurrentValue > $LastValue);
        if ($BMIValues{$CurrentValue} eq "BMI") {
          $GraphRow[$x].=$SpotMark;
        }
        $LastValue=$CurrentValue + 1;
      }
      $GraphRow[$x].="<tr><td width=100% height=".((scalar 110 - $LastValue) * 2)."></td></tr>\n";

      $GraphRow[$x].="</table></td>";
      
      #Save this date for next go round
      $LastDate=$GraphEventDate;
    }

    #Print the report header
    print "
    <a name=GRAPHBMI></a>
    <table bgcolor=".$ReportColour."  border=0 cellspacing=0 cellpadding=2 width=683>
    <tr>
      <td>
      <table  bgcolor=white border=0 cellspacing=0 cellpadding=0 width=100%>
        <tr>
        <th bgcolor=".$ReportColour.">".$LocGraphBMIHeader."</th>
        </tr>
        <tr>
        <td height=12>
        </td>
        </tr>
        <tr>
        <td>
          <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=640 height=240>
          <tr>
            <td width=34 height=221 bgcolor=white rowspan=3>
            <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=100% height=100%>
              <tr>
              <td valign=top align=right height=50%>".$GraphMax." ".$BMIUnits."</td>
              </tr>
              <tr>
              <td valign=bottom align=right height=50%>".$GraphMin."</td>
              </tr>
            </table>
            </td>
            <td width=6 height=1 bgcolor=grey colspan=2></td>
            <td width=600 height=220 bgcolor=white colspan=2 rowspan=2>
            <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=600 height=220>";

            #Print each log entry from the Hash tables, extract each one by the timestamp

            #Print row by row
            $blankspace= 150 - scalar $i;

            #Start new row
            print "
                                <tr width=600 height=220 valign=bottom>";

            #Print oldest to newest (left to right) so it aligns with the LH axis
            for ($x = $i; $x >= 1; $x--) {
              print $GraphRow[$x];
            }

            print "<td width=".$blankspace."></td>";
            #End row
            print "</tr>";
            
            #End the table
            print "
            </table>
            </td>
          </tr>
          <tr>
            <td width=5 height=219 bgcolor=white>
              <table bgcolor=white border=0 cellspacing=0 cellpadding=0 height=219>
                <tr><td colspan=2 width=5 height=21></td></tr>
                <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=21></td></tr>
                <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=21></td></tr>
                <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=21></td></tr>
                <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=21></td></tr>
                <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=21></td></tr>
                <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=21></td></tr>
                <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=21></td></tr>
                <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=21></td></tr>
                <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
                <tr><td colspan=2 width=5 height=21></td></tr>
              </table>
            </td>
            <td width=1 height=219 bgcolor=grey></td>
          </tr>
          <tr>
            <td width=606 height=1 bgcolor=grey colspan=4></td>
          </tr>
          <tr>
            <td width=34 height=5 bgcolor=white></td>
            <td width=5 height=5 bgcolor=white></td>
            <td width=1 height=5 bgcolor=grey></td>
            <td width=599 height=5 bgcolor=white></td>
            <td width=1 height=5 bgcolor=grey></td>
          </tr>
          <tr>
            <td width=39 height=14 bgcolor=white colspan=2></td>
            <td width=601 height=14 bgcolor=white colspan=3>
            <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=100% height=100%>
              <tr>
              <td valign=top align=left width=50%>".$TimeMin."</td>
              <td valign=top align=right width=50%>".$TimeMax."</td>
              </tr>
            </table>
            </td>
          </tr>
          </table>
        </td>
        </tr>".$NumReadingsText."
        <tr>
        <td>
          <table bgcolor=white border=0 cellspacing=0 cellpadding=0>
          <tr><td colspan=4><b>".$LocGraph0."</b></td></tr>
          <tr>
            <td width=5></td>
            <td width=10 bgcolor=lightcyan></td>
            <td>".$LocGraph4."</td><td width=5></td>
          </tr>
          </table
        </td>
        </tr>
      </table>
      </td>
    </tr>
    </table>
    <p>&nbsp;
    ";
  }
}


#==============================================================================
#
# SUBROUTINES - PRINT REPORTS - GRAPH SUMMARY
#
#==============================================================================

#=======================================
# Report Generation Subroutines - Graph Summary
#=======================================
sub PrintGraphSummary() {

  #Establish variables for storing glucose data
  my $AvgGlucose;
  my $NumEvents;

  #Setup graph defaults
  if ($GlucoseUnits eq "mmol/L") {
    $GraphCriticalHigh=sprintf("%0.0f", scalar $GlucoseLevels{"CriticalHigh"} * 10);
    $GraphTargetHigh=sprintf("%0.0f", scalar $GlucoseLevels{"TargetHigh"} * 10);
    $GraphTargetLow=sprintf("%0.0f", scalar $GlucoseLevels{"TargetLow"} * 10);
    $GraphCriticalLow=sprintf("%0.0f", scalar $GlucoseLevels{"CriticalLow"} * 10);
  }
  else {
    $GraphCriticalHigh=sprintf("%0.0f", (scalar $GlucoseLevels{"CriticalHigh"} * 0.5));
    $GraphTargetHigh=sprintf("%0.0f", (scalar $GlucoseLevels{"TargetHigh"} * 0.5));
    $GraphTargetLow=sprintf("%0.0f", (scalar $GlucoseLevels{"TargetLow"} * 0.5));
    $GraphCriticalLow=sprintf("%0.0f", (scalar $GlucoseLevels{"CriticalLow"} * 0.5));
  }
  
  #Get reverse order of glucose timestamps
  (@GraphTimestamps)=sort InverseSort (keys %Glucose);
  
  $i=@Periods;  #Number of columns populated
  undef %SumGraphGlucose;
  undef %AvgGraphGlucose;
  undef %NumGraphGlucose;
  undef %MaxGraphGlucose;
  undef %MinGraphGlucose;
  #undef %SQSumGraphGlucose;
  #undef %RMSGraphGlucose;
  undef @$EventPeriod;
  foreach (@GraphTimestamps) {
    $EventPeriod=$Period{$_};
    $SumGraphGlucose{$EventPeriod} += $Glucose{$_};
    $NumGraphGlucose{$EventPeriod}++;
    $AvgGraphGlucose{$EventPeriod} = $SumGraphGlucose{$EventPeriod}/$NumGraphGlucose{$EventPeriod};
    $MaxGraphGlucose{$EventPeriod} = $Glucose{$_} if ((!$MaxGraphGlucose{$EventPeriod}) || ($MaxGraphGlucose{$EventPeriod} < $Glucose{$_}));
    $MinGraphGlucose{$EventPeriod} = $Glucose{$_} if ((!$MinGraphGlucose{$EventPeriod}) || ($MinGraphGlucose{$EventPeriod} > $Glucose{$_}));
    #$SQSumGraphGlucose{$EventPeriod} += ($Glucose{$_}**2);
    #$RMSGraphGlucose{$EventPeriod} = sqrt ($SQSumGraphGlucose{$EventPeriod}/$NumGraphGlucose{$EventPeriod});
    push @$EventPeriod, $Glucose{$_};
  }
  
  #Print the report header
  print "
  <a name=GRAPHSUMMARY></a>
  <table bgcolor=".$ReportColour."  border=0 cellspacing=0 cellpadding=2 width=683>
    <tr>
      <td>
        <table  bgcolor=white border=0 cellspacing=0 cellpadding=0 width=100%>
          <tr>
            <th bgcolor=".$ReportColour.">".$LocGraphSummaryHeader."</th>
          </tr>
          <tr>
            <td height=12>
            </td>
          </tr>
          <tr>
            <td>
              <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=640 height=240>
                <tr>
                  <td width=34 height=221 bgcolor=white rowspan=3>
                  ".$GraphYscale."
                  </td>
                  <td width=6 height=1 bgcolor=grey colspan=2></td>
                  <td width=600 height=222 bgcolor=white colspan=2 rowspan=2 valign=middle>
                    <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=600 height=220>
                      <tr>";

  #Print each log entry from the Hash tables, extract each one by the timestamp
  undef @SummaryPeriods;
  foreach (@Periods) {
    if ($NumGraphGlucose{$_}) {
      push @SummaryPeriods, $_;
    }
  }
  
  #Print the summary report for each period, sorted alphabetically
  foreach (@SummaryPeriods) {
    $ReportPeriod=$_;
        
    undef $SDval;
    foreach (@$ReportPeriod){
      $SDval += ($_ - $AvgGraphGlucose{$ReportPeriod})**2;
    }

    # [v1.3.1 bug fixed] $EventPeriod is now $ReportPeriod
    $SDval= sqrt ($SDval/$NumGraphGlucose{$ReportPeriod}) if ($NumGraphGlucose{$ReportPeriod});

    @PeriodValues=($AvgGraphGlucose{$ReportPeriod}, $AvgGraphGlucose{$ReportPeriod} + $SDval, $AvgGraphGlucose{$ReportPeriod} - $SDval);
    
    foreach ($i = 0; $i <= 2; $i++) { 
      if ($GlucoseUnits eq "mmol/L") {
        if (scalar $PeriodValues[$i] >= 22.0) {
          $PeriodValues[$i]="220";
        }
        elsif (scalar $PeriodValues[$i] <= 0.0) {
          $PeriodValues[$i]="1";
        }
        else {
          $PeriodValues[$i]=sprintf("%0.0f", scalar $PeriodValues[$i] * 10);
        }
      }
      else {
        if (scalar $PeriodValues[$i] >= 440) {
          $PeriodValues[$i]="220";
        }
        elsif (scalar $PeriodValues[$i] <= 0.0) {
          $PeriodValues[$i]="1";
        }
        else {
          $PeriodValues[$i]=sprintf("%0.0f", scalar $PeriodValues[$i] * 0.5);
        }
      }
    }
    
    if ($SDval > 0) {
      $Height1=220 - $PeriodValues[1];
      $Height2=1;
      $Height3=$PeriodValues[1] - $PeriodValues[0] - 5;
      $Height4=9;
      $Height5=$PeriodValues[0] - $PeriodValues[2] - 5;
      $Height6=1;
      $Height7=$PeriodValues[2] - 1;
    }
    else {
      $Height1=220 - $PeriodValues[1] - 5;
      $Height2=1;
      $Height3=0;
      $Height4=9;
      $Height5=0;
      $Height6=1;
      $Height7=$PeriodValues[2] - 6;
    }
    
    if (scalar $AvgGraphGlucose{$ReportPeriod} < $GlucoseLevels{"CriticalLow"}) {
      $AvgColor=$GlucoseColors{"VeryLow"};
    }
    elsif (scalar $AvgGraphGlucose{$ReportPeriod} < $GlucoseLevels{"TargetLow"}) {
      $AvgColor=$GlucoseColors{"Low"};
    }
    elsif (scalar $AvgGraphGlucose{$ReportPeriod} < $GlucoseLevels{"TargetHigh"}) {
      $AvgColor=$GlucoseColors{"Target"};
    }
    elsif (scalar $AvgGraphGlucose{$ReportPeriod} < $GlucoseLevels{"CriticalHigh"}) {
      $AvgColor=$GlucoseColors{"High"};
    }
    else {
      $AvgColor=$GlucoseColors{"VeryHigh"};
    }

    print "
          <td width=".scalar (600/@SummaryPeriods)." title=\"".$LocAverage."=".sprintf("%0.1f", $AvgGraphGlucose{$ReportPeriod})."\n".$LocMaximum."=".sprintf("%0.1f", $MaxGraphGlucose{$ReportPeriod})."\n".$LocMinimum."=".sprintf("%0.1f", $MinGraphGlucose{$ReportPeriod})."\n".$LocGraphSummary1."=".sprintf("%0.1f", $SDval)."\">
            <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=".scalar (600/@SummaryPeriods)." height=220>
              <tr>
                <td height=220>
                  <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=".scalar ((600/@SummaryPeriods - 9)/2 - 5)." height=220>
                    <tr><td width=".scalar ((600/@SummaryPeriods - 9)/2 - 5)." height=".scalar 220 - $GraphCriticalHigh."></td></tr>
                    <tr><td height=1 bgcolor=".$GlucoseColors{"VeryHigh"}."></td> </tr>
                    <tr><td height=".scalar ($GraphCriticalHigh - $GraphTargetHigh - 1)."></td></tr>
                    <tr><td height=1 bgcolor=".$GlucoseColors{"Target"}."></td> </tr>
                    <tr><td height=".scalar ($GraphTargetHigh - $GraphTargetLow - 1)."></td></tr>
                    <tr><td height=1 bgcolor=".$GlucoseColors{"Target"}."></td> </tr>
                    <tr><td height=".scalar ($GraphTargetLow - $GraphCriticalLow - 1)."></td></tr>
                    <tr><td height=1 bgcolor=".$GlucoseColors{"VeryHigh"}."></td> </tr>
                    <tr><td height=".scalar ($GraphCriticalLow - 1)."></td></tr>
                  </table>
                </td>
                <td width=5 height=220></td>
                <td height=220>
                  <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=9 height=220>
                   <tr><td height=".$Height1."></td></tr>
                  <tr><td height=".$Height2." bgcolor=gray></td> </tr>
                  <tr>
                <td>
                  <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center  height=".$Height3." width=1>
                    <tr><td width=1 bgcolor=gray></td></tr>
                  </table>
                </td>
              </tr>
              <tr>
                <td>
                  <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=9  height=".$Height4.">
                    <tr><td><table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center height=1><tr><td bgcolor=".$AvgColor." width=1></td></tr></table></td></tr>
                    <tr><td><table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center height=1><tr><td bgcolor=".$AvgColor." width=3></td></tr></table></td></tr>
                    <tr><td><table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center height=1><tr><td bgcolor=".$AvgColor." width=5></td></tr></table></td></tr>
                    <tr><td><table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center height=1><tr><td bgcolor=".$AvgColor." width=7></td></tr></table></td></tr>
                    <tr><td><table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center height=1><tr><td bgcolor=".$AvgColor." width=9></td></tr></table></td></tr>
                    <tr><td><table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center height=1><tr><td bgcolor=".$AvgColor." width=7></td></tr></table></td></tr>
                    <tr><td><table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center height=1><tr><td bgcolor=".$AvgColor." width=5></td></tr></table></td></tr>
                    <tr><td><table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center height=1><tr><td bgcolor=".$AvgColor." width=3></td></tr></table></td></tr>
                    <tr><td><table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center height=1><tr><td bgcolor=".$AvgColor." width=1></td></tr></table></td></tr>
                  </table>
                </td>
              </tr>
              <tr>
                <td>
                  <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center  height=".$Height5." width=1>
                    <tr><td width=1 bgcolor=gray></td></tr>
                  </table>
                </td>
              </tr>
              <tr><td height=".$Height6." bgcolor=gray></td></tr>
              <tr><td height=".$Height7."></td></tr>
            </table>
          </td>
          <td width=5 height=220></td>
          <td height=220>
            <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=".scalar ((600/@SummaryPeriods - 9)/2 - 5)." height=220>
              <tr><td width=".scalar ((600/@SummaryPeriods - 9)/2 - 5)." height=".scalar 220 - $GraphCriticalHigh."></td></tr>
                <tr><td height=1 bgcolor=".$GlucoseColors{"VeryHigh"}."></td> </tr>
                <tr><td height=".scalar ($GraphCriticalHigh - $GraphTargetHigh - 1)."></td></tr>
                <tr><td height=1 bgcolor=".$GlucoseColors{"Target"}."></td> </tr>
                <tr><td height=".scalar ($GraphTargetHigh - $GraphTargetLow - 1)."></td></tr>
                <tr><td height=1 bgcolor=".$GlucoseColors{"Target"}."></td> </tr>
                <tr><td height=".scalar ($GraphTargetLow - $GraphCriticalLow - 1)."></td></tr>
                <tr><td height=1 bgcolor=".$GlucoseColors{"VeryHigh"}."></td> </tr>
                <tr><td height=".scalar ($GraphCriticalLow - 1)."></td></tr>
              </table>
            </td>
          </tr>
        </table>
      </td>";
  }
  
  #End the table
  print "
                      </tr>
                    </table>
                  </td>
                </tr>
                <tr>
          <td width=5 height=219 bgcolor=white>
            <table bgcolor=white border=0 cellspacing=0 cellpadding=0 height=219>
              <tr><td colspan=2 width=3 height=19></td></tr>
              <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=19></td></tr>
              <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=19></td></tr>
              <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=19></td></tr>
              <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=19></td></tr>
              <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=19></td></tr>
              <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=19></td></tr>
              <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=19></td></tr>
              <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=19></td></tr>
              <tr><td colspan=2 width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=19></td></tr>
              <tr><td width=2 height=1></td><td width=3 height=1 bgcolor=grey></td></tr>
              <tr><td colspan=2 width=5 height=19></td></tr>
            </table>
          </td>
          <td width=1 height=219 bgcolor=grey></td>
                </tr>
                <tr>
                  <td width=606 height=1 bgcolor=grey colspan=4></td>
                </tr>
                <tr>
                  <td width=34 height=5 bgcolor=white></td>
                  <td width=5 height=5 bgcolor=white></td>
                  <td width=1 height=5 bgcolor=grey></td>
                  <td width=601 height=14 bgcolor=white colspan=2>
                    <table bgcolor=white border=0 cellspacing=0 cellpadding=0 align=center width=100% height=100%>
                      <tr>";

  foreach (@SummaryPeriods) {
    $ReportPeriod=$_;
    $ReportPeriod=~ s/ /<br>/g;
    
    print "
                        <td width=".scalar (600/@SummaryPeriods)." align=center>$ReportPeriod
                        </td>
                        <td width=1 bgcolor=gray>
                        </td>";
  }
  

  print "
                      </tr>
                    </table>
                  </td>
                </tr>
              </table>
            </td>
          </tr>
          <tr>
            <td>
              <table bgcolor=white border=0 cellspacing=0 cellpadding=0>
                <tr><td colspan=6><b>".$LocGraph0."</b></td></tr>
                <tr>
                  <td width=5></td>
                  <td valign=middle width=10>
                    <table bgcolor=white border=0 cellspacing=0 cellpadding=0>
                      <tr><td width=10 height=2 bgcolor=".$GlucoseColors{"VeryHigh"}." width=100%></td></tr></table>
                  </td>
                  <td>".$LocGraph1.",</td>
                  <td width=5></td>
                  <td valign=middle width=10>
                    <table bgcolor=white border=0 cellspacing=0 cellpadding=0>
                      <tr><td width=10 height=2 bgcolor=".$GlucoseColors{"Target"}." width=100%></td></tr></table>
                  </td>
                  <td>".$LocGraph2."</td>
                </tr>
              </table>
              <table bgcolor=white border=0 cellspacing=0 cellpadding=0>
                <tr>
                  <td width=5></td>
                  <td colspan=5>".$LocGraphSummary0."</td>
                </tr>
              </table>
            </td>
          </tr>
        </table>
      </td>
    </tr>
  </table>
  <p>&nbsp;
  ";

  undef $ReportPeriod;
}



#==============================================================================
#
# SUBROUTINES - PRINT REPORTS - NOTES
#
#==============================================================================

#=======================================
# Report Generation Subroutines - Notes
#=======================================
sub PrintNotes() {

  print "
  <a name=NOTES></a>";

  #Start report table, with Title, and column headers
  print "
  <table bgcolor=".$ReportColour."  border=0 cellspacing=0 cellpadding=2 width=683>
    <tr>
      <td>
        <table  bgcolor=white border=0 cellspacing=0 cellpadding=0 width=100%>
          <tr>
            <th colspan=16 bgcolor=".$ReportColour.">".$LocNotesHeader."</th>
          </tr>
          <tr>
            <td colspan=16 bgcolor=".$HeadingColour." height=1></td>
          </tr>
          <tr bgcolor=".$ReportColour.">
            <td width=50 style=\"font-size: 10px; color: ".$HeadingColour."; text-decoration:bold;\">".$LocNotesCol1."</td>
            <td bgcolor=".$HeadingColour." width=1></td>
            <td align=left width=".(30 + $TimeAdjustWidth)." style=\"font-size: 10px; color: ".$HeadingColour."; text-decoration:bold;\">".$LocNotesCol2."</td>
            <td bgcolor=".$HeadingColour." width=1></td>
            <td align=left width=102 style=\"font-size: 10px; color: ".$HeadingColour."; text-decoration:bold;\">".$LocNotesCol3."</td>
            <td bgcolor=".$HeadingColour." width=1></td>
            <td width=".(492 - $TimeAdjustWidth)." style=\"font-size: 10px; color: ".$HeadingColour."; text-decoration:bold;\">".$LocNotesCol4."</td>
          </tr>";


  #Print each log entry from the Hash tables, extract each one by the timestamp
  undef $ThisDate;
  undef $ThisPeriod;
  undef $myDay;
  undef $myMonth;
  undef $myYear;
  undef $myLastDay;
  undef $myLastMonth;
  undef $myLastYear;
  undef $DayName;
  undef $WeekDate;
  $ThatTime=0;
  foreach (@Timestamp) {
 
    $HasData="NO";
    $HasData="YES" if ($Glucose{$_});
    $HasData="YES" if ($Insulin{$_});
    $HasData="YES" if ($Carbs{$_});
    $HasData="YES" if ($InsulinB{$_});
    $HasData="YES" if ($Medicine{$_});
    $HasData="YES" if ($Ketones{$_});
    $HasData="YES" if ($Weight{$_});
    $HasData="YES" if ($Height{$_});
    $HasData="YES" if ($HbA1c{$_});
    
    #Only process data is present
    if ($HasData eq "YES") {
      next;
    }


    #Only process required period, if printing a summary report
    if ($ReportPeriod) {
      next if ($Period{$_} ne $ReportPeriod);
    }
    
    $WeekDate="";
    $ThisWeek=$Date{$_} if   (!$ThisWeek);
    #Print a break row if the date has changed
    if ($ThisDate ne $Date{$_}) {
  
      ($myDay,$myMonth,$myYear)=split $DateSep , $Date{$_} if ($DateFormat eq "day/month/year");
      ($myMonth,$myDay,$myYear)=split $DateSep , $Date{$_} if ($DateFormat eq "month/day/year");
      $ThisTime=timelocal("","","", scalar $myDay, scalar $myMonth - 1, scalar $myYear - 1900);
      $DayName=$Day{$_};
      
      ($myLastDay,$myLastMonth,$myLastYear)=split $DateSep , $ThisWeek if ($DateFormat eq "day/month/year");
      ($myLastMonth,$myLastDay,$myLastYear)=split $DateSep , $ThisWeek if ($DateFormat eq "month/day/year");
      $ThatTime=timelocal("","","", scalar $myLastDay, scalar $myLastMonth - 1, scalar $myLastYear - 1900) if ($ThisDate);
      
      if ($DayName eq "Mon") {
        $WeekDate=$Date{$_};
        $ThisWeek=$Date{$_};
      }
      elsif ((scalar $ThisTime - scalar $ThatTime) >= 604800) {
        $ThatTime=scalar $ThisTime if ($DayName eq $DaysOfWeek{"Mon"});
        $ThatTime=scalar $ThisTime - (86400 * 1) if ($DayName eq $DaysOfWeek{"Tue"});
        $ThatTime=scalar $ThisTime - (86400 * 2) if ($DayName eq $DaysOfWeek{"Wed"});
        $ThatTime=scalar $ThisTime - (86400 * 3) if ($DayName eq $DaysOfWeek{"Thu"});
        $ThatTime=scalar $ThisTime - (86400 * 4) if ($DayName eq $DaysOfWeek{"Fri"});
        $ThatTime=scalar $ThisTime - (86400 * 5) if ($DayName eq $DaysOfWeek{"Sat"});
        $ThatTime=scalar $ThisTime - (86400 * 6) if ($DayName eq $DaysOfWeek{"Sun"});
        (@TWelements)=localtime($ThatTime);
        $TWDay=scalar $TWelements[3];
        $TWMonth=1 + scalar $TWelements[4];
        $TWDay="0".$TWDay if ($TWDay <= 9);
        $TWMonth="0".$TWMonth if ($TWMonth <= 9);
        $TWYear=1900 + scalar $TWelements[5];
        $ThisWeek=$TWDay.$DateSep.$TWMonth.$DateSep.$TWYear if ($DateFormat eq "day/month/year");
        $ThisWeek=$TWMonth.$DateSep.$TWDay.$DateSep.$TWYear if ($DateFormat eq "month/day/year");
        $WeekDate=$ThisWeek;
      }
      
      $ThisPeriod=$Period{$_};
      if ($WeekDate) {
        print "
          <tr>
            <td colspan=7 bgcolor=".$ReportColour." height=1></td>
          </tr>";
      }
      else {
        print "
          <tr>
            <td colspan=7 bgcolor=gray height=1></td>
          </tr>";
      }
      $ThisDate="<table border=0 cellpadding=0 cellspacing=0><tr><td width=25>".$DayName."</td><td>(".$myDay.")</td></tr></table>";
    }
    else {
      
      #Don't print row separators when the date or period hasn't changed
      $ThisDate="";
      $myDay="";
      $myMonth="";
      $myYear="";
      $DayName="";
      if ($ThisPeriod ne $Period{$_}) {
        $ThisPeriod=$Period{$_};
        $ThisPeriodColor="lightgrey";
      }
      else {
        $ThisPeriod="";
        $ThisPeriodColor="white";
      }
      print "
          <tr>
            <td height=1></td>
            <td bgcolor=lightgrey height=1></td>
            <td bgcolor=".$ThisPeriodColor." height=1></td>
            <td colspan=4 bgcolor=lightgrey height=1></td>
          </tr>";
    }
    
    print "
          <tr>
            <td colspan=7 bgcolor=".$ReportColour." height=1></td>
          </tr>
          <tr>
            <td colspan=7 bgcolor=gray align=left style=\"color:".$HeadingColour.";\"><b>".$LocWeekStarting." ".$WeekDate."</b></td>
          </tr>" if ($WeekDate);

  
    $AddNotableColour="";
    $AddNotableColour=" bgcolor=LIGHTGOLDENRODYELLOW" if ($Notable{$_});
    $Notes=$Notes{$_};
    $Notes="<font style=\"font-family: arial; color: black; font-size: 11px;\"><b>".$Notes."</b></font>" if ($Notes =~ m/HbA1c/ );
  
    #Print part of the entry
    print "
          <tr valign=top".$AddNotableColour.">
            <td>".$ThisDate."</td>
            <td bgcolor=lightgrey width=1></td>
            <td align=right style=\"color: darkslategray;\">".$Time{$_}."</td>
            ";
   
    #Print end of the entry
    print "</td>
            <td bgcolor=lightgrey width=1></td>
            <td width=102><font color=navy><b>".$Eventname{$_}."</b></font></td>
            <td bgcolor=lightgrey width=1></td>
            <td>".$Notes."</td>
          <tr>";

    #Set values for use with next log entry
    $ThisDate=$Date{$_};
    $ThisPeriod=$Period{$_};
  }
  #All entries are now printed
    
  #Close out the table
  print "
        </table>
      </td>
    </tr>
  </table>
  <p>&nbsp;
  ";

}



#==============================================================================
#
# SUBROUTINES - PRINT REPORTS - DETAILED & SUMMARY
#
#==============================================================================

#=======================================
# Report Generation Subroutines - List
#=======================================
sub PrintReport() {

  #Establish variables for storing glucose data
  my $MaxGlucose;
  my $MinGlucose;
  my $AvgGlucose;
  my $NumGlucose;

  my $ShowOtherMed1="NO";
  my $ShowOtherMed2="NO";
  my $NumCols=12;
  if ($opt_i || $opt_m) {
    $ShowOtherMed1="YES";
    $NumCols+=4;
  }
  if ($opt_h || $opt_c) {
    $ShowOtherMed2="YES";
    $NumCols+=4;
  }

  #Print the right header, if ReportPeriod the we are printing a summary  
  if ($ReportPeriod) {
    $ReportTitle=$LocSummaryHeader."&nbsp;&nbsp;<font color=".$GlucoseColors{"VeryHigh"}.">".$ReportPeriod."</font>";
    print "
  <a name=SUMMARY></a>" if (!$FirstPeriod);
  $FirstPeriod="NO";
  }
  else {
    $ReportTitle=$LocDetailedHeader;
    print "
  <a name=DETAILED></a>";
  }

  #Start report table, with Title, and column headers
  print "
  <table bgcolor=".$ReportColour."  border=0 cellspacing=0 cellpadding=2 width=683>
    <tr>
      <td>
        <table  bgcolor=white border=0 cellspacing=0 cellpadding=0>
          <tr>
            <th colspan=".$NumCols." bgcolor=".$ReportColour.">".$ReportTitle."</th>
          </tr>
          <tr>
            <td colspan=".$NumCols." bgcolor=".$HeadingColour." height=1></td>
          </tr>
          <tr bgcolor=".$ReportColour.">
            <td width=50 style=\"font-size: 10px; color: ".$HeadingColour."; text-decoration:bold;\">".$LocReportCol1."</td>
            <td bgcolor=".$HeadingColour." width=1></td>
            <td width=115 style=\"font-size: 10px; color: ".$HeadingColour."; text-decoration:bold;\">".$LocReportCol2."</td>
            <td bgcolor=".$HeadingColour." width=1></td>
            <td align=left width=".(30 + $TimeAdjustWidth)." style=\"font-size: 10px; color: ".$HeadingColour."; text-decoration:bold;\">".$LocReportCol3."</td>
            <td bgcolor=".$HeadingColour." width=1></td>
            <td colspan=2 align=left width=102 style=\"font-size: 10px; color: ".$HeadingColour."; text-decoration:bold;\">".$LocReportCol4."</td>
            <td bgcolor=".$HeadingColour." width=1></td>";
  print "
            <td colspan=3 align=left width=115 style=\"font-size: 10px; color: ".$HeadingColour."; text-decoration:bold;\">".$LocReportCol5."</td>
            <td bgcolor=".$HeadingColour." width=1></td>" if ($ShowOtherMed1 eq "YES");
  print "
            <td colspan=3 align=left width=93 style=\"font-size: 10px; color: ".$HeadingColour."; text-decoration:bold;\">".$LocReportCol5a."</td>
            <td bgcolor=".$HeadingColour." width=1></td>" if ($ShowOtherMed2 eq "YES");
  print "
            <td width=5></td>
            <td width=1></td>
            <td width=".(371 - $TimeAdjustWidth - $MedsAdjustWidth - $InsAdjustWidth)." style=\"font-size: 10px; color: ".$HeadingColour."; text-decoration:bold;\">".$LocReportCol6."</td>
          </tr>";


  #Print each log entry from the Hash tables, extract each one by the timestamp
  undef $ThisDate;
  undef $ThisPeriod;
  undef $myDay;
  undef $myMonth;
  undef $myYear;
  undef $myLastDay;
  undef $myLastMonth;
  undef $myLastYear;
  undef $DayName;
  undef $WeekDate;
  $ThatTime=0;
  foreach (@Timestamp) {
    
    $HasData="NO";
    $HasData="YES" if ($Glucose{$_});
    $HasData="YES" if ($Insulin{$_} && ($opt_i));
    $HasData="YES" if ($Carbs{$_} && ($opt_c));
    $HasData="YES" if ($InsulinB{$_} && ($opt_i));
    $HasData="YES" if ($Medicine{$_} && ($opt_m));
    $HasData="YES" if ($Ketones{$_});
    $HasData="YES" if ($Weight{$_} && ($opt_h));
    $HasData="YES" if ($Height{$_} && ($opt_h));
    $HasData="YES" if ($HbA1c{$_} && ($opt_h));
    
    #Only process data is present
    if ($HasData eq "NO") {
      next;
    }


    #Only process required period, if printing a summary report
    if ($ReportPeriod) {
      next if ($Period{$_} ne $ReportPeriod);
    }
    
    $WeekDate="";
    $ThisWeek=$Date{$_} if   (!$ThisWeek);
    #Print a break row if the date has changed
    if ($ThisDate ne $Date{$_}) {
  
      ($myDay,$myMonth,$myYear)=split $DateSep , $Date{$_} if ($DateFormat eq "day/month/year");
      ($myMonth,$myDay,$myYear)=split $DateSep , $Date{$_} if ($DateFormat eq "month/day/year");
      $ThisTime=timelocal("","","", scalar $myDay, scalar $myMonth - 1, scalar $myYear - 1900);
      $DayName=$Day{$_};
      
      ($myLastDay,$myLastMonth,$myLastYear)=split $DateSep , $ThisWeek if ($DateFormat eq "day/month/year");
      ($myLastMonth,$myLastDay,$myLastYear)=split $DateSep , $ThisWeek if ($DateFormat eq "month/day/year");
      $ThatTime=timelocal("","","", scalar $myLastDay, scalar $myLastMonth - 1, scalar $myLastYear - 1900) if ($ThisDate);
      
      if ($DayName eq "Mon") {
        $WeekDate=$Date{$_};
        $ThisWeek=$Date{$_};
      }
      elsif ((scalar $ThisTime - scalar $ThatTime) >= 604800) {
        $ThatTime=scalar $ThisTime if ($DayName eq $DaysOfWeek{"Mon"});
        $ThatTime=scalar $ThisTime - (86400 * 1) if ($DayName eq $DaysOfWeek{"Tue"});
        $ThatTime=scalar $ThisTime - (86400 * 2) if ($DayName eq $DaysOfWeek{"Wed"});
        $ThatTime=scalar $ThisTime - (86400 * 3) if ($DayName eq $DaysOfWeek{"Thu"});
        $ThatTime=scalar $ThisTime - (86400 * 4) if ($DayName eq $DaysOfWeek{"Fri"});
        $ThatTime=scalar $ThisTime - (86400 * 5) if ($DayName eq $DaysOfWeek{"Sat"});
        $ThatTime=scalar $ThisTime - (86400 * 6) if ($DayName eq $DaysOfWeek{"Sun"});
        (@TWelements)=localtime($ThatTime);
        $TWDay=scalar $TWelements[3];
        $TWMonth=1 + scalar $TWelements[4];
        $TWDay="0".$TWDay if ($TWDay <= 9);
        $TWMonth="0".$TWMonth if ($TWMonth <= 9);
        $TWYear=1900 + scalar $TWelements[5];
        $ThisWeek=$TWDay.$DateSep.$TWMonth.$DateSep.$TWYear if ($DateFormat eq "day/month/year");
        $ThisWeek=$TWMonth.$DateSep.$TWDay.$DateSep.$TWYear if ($DateFormat eq "month/day/year");
        $WeekDate=$ThisWeek;
      }
      
      $ThisPeriod=$Period{$_};
      if ($WeekDate) {
        print "
          <tr>
            <td colspan=".$NumCols." bgcolor=".$ReportColour." height=1></td>
          </tr>";
      }
      else {
        print "
          <tr>
            <td colspan=".$NumCols." bgcolor=gray height=1></td>
          </tr>";
      }
      $ThisDate="<table border=0 cellpadding=0 cellspacing=0><tr><td width=25>".$DayName."</td><td>(".$myDay.")</td></tr></table>";
    }
    else {
      
      #Don't print row separators when the date or period hasn't changed
      $ThisDate="";
      $myDay="";
      $myMonth="";
      $myYear="";
      $DayName="";
      if ($ThisPeriod ne $Period{$_}) {
        $ThisPeriod=$Period{$_};
        $ThisPeriodColor="lightgrey";
      }
      else {
        $ThisPeriod="";
        $ThisPeriodColor="white";
      }
      print "
          <tr>
            <td height=1></td>
            <td bgcolor=lightgrey height=1></td>
            <td bgcolor=".$ThisPeriodColor." height=1></td>
            <td colspan=".(scalar $NumCols - 3)." bgcolor=lightgrey height=1></td>
          </tr>";
    }
    
    #Determine the colour of the of the glucose field, using the preset values
    if (scalar $Glucose{$_} < $GlucoseLevels{"CriticalLow"}) {
      $Gcolor=$GlucoseColors{"VeryLow"};
    }
    elsif (scalar $Glucose{$_} < $GlucoseLevels{"TargetLow"}) {
      $Gcolor=$GlucoseColors{"Low"};
    }
    elsif (scalar $Glucose{$_} < $GlucoseLevels{"TargetHigh"}) {
      $Gcolor=$GlucoseColors{"Target"};
    }
    elsif (scalar $Glucose{$_} < $GlucoseLevels{"CriticalHigh"}) {
      $Gcolor=$GlucoseColors{"High"};
    }
    else {
      $Gcolor=$GlucoseColors{"VeryHigh"};
    }
    
    print "
          <tr>
            <td colspan=".$NumCols." bgcolor=".$ReportColour." height=1></td>
          </tr>
          <tr>
            <td colspan=".$NumCols." bgcolor=gray align=left style=\"color:".$HeadingColour.";\"><b>".$LocWeekStarting." ".$WeekDate."</b></td>
          </tr>" if ($WeekDate);

  
    $AddNotableColour="";
    $AddNotableColour=" bgcolor=LIGHTGOLDENRODYELLOW" if ($Notable{$_});
  
    #Print part of the entry
    print "
          <tr valign=top".$AddNotableColour.">
            <td>".$ThisDate."</td>
            <td bgcolor=lightgrey width=1></td>
            <td>".$ThisPeriod."</td><td bgcolor=lightgrey width=1></td>
            <td align=right style=\"color: darkslategray;\">".$Time{$_}."</td>
            <td bgcolor=lightgrey width=1></td>
            <td width=26 align=right style=\"color:".$Gcolor.";\">".$Glucose{$_}."</td>
            <td width=76>";
    
      #Print a graph of the glucose, with coloured levels
    if ($Glucose{$_}) {
      ($WholeGlucose,$Pad)=split /\./ , $Glucose{$_};
    
      print "
              <table border=0 cellspacing=0 cellpadding=0>
                <tr><td>&nbsp;</td>";
    
      for ($Scale = 1; $Scale <= 24; $Scale++) {
        if ($GlucoseUnits eq "mg/dL") {
          $TestScale=scalar $Scale*$MGperL;
        }
        else {
          $TestScale=$Scale;
        }

        if (scalar $TestScale < $GlucoseLevels{"CriticalLow"}) {
          $Scolor=$GlucoseColors{"VeryLow"};
        }
        elsif (scalar $TestScale < $GlucoseLevels{"TargetLow"}) {
          $Scolor=$GlucoseColors{"Low"};
        }
        elsif (scalar $TestScale < $GlucoseLevels{"TargetHigh"}) {
          $Scolor=$GlucoseColors{"Target"};
        }
        elsif (scalar $TestScale < $GlucoseLevels{"CriticalHigh"}) {
          $Scolor=$GlucoseColors{"High"};
        }
        else {
          $Scolor=$GlucoseColors{"VeryHigh"};
        }

        if (scalar $TestScale > $WholeGlucose) {
          last;
        }

        print "
                  <td width=1 bgcolor=\"".$Scolor."\">&nbsp;</td>";
      }
    
      print "
                </tr>
              </table>";
    
    }
    
    $OtherMeds="";
    $OtherMedsNames="";
  
    if ($Insulin{$_} && ($opt_i)) {
      $OtherMeds.="<font color=navy>".$Insulin{$_}."</font>";
      if ($InsType{$_}) {
        $OtherMedsNames.="<font color=navy>".$InsType{$_}."</font>";
      }
      else {
        $OtherMedsNames.="<font color=navy>".$LocInsulinA."</font>";
      }
    }
    if ($InsulinB{$_} && ($opt_i)) {
      if ($OtherMeds) {
        $OtherMeds.="<br>";
        $OtherMedsNames.="<br>";
      }
      $OtherMeds.="<font color=teal>".$InsulinB{$_}."</font>";
      if ($InsBType{$_}) {
        $OtherMedsNames.="<font color=teal>".$InsBType{$_}."</font>";
      }
      else {
        $OtherMedsNames.="<font color=teal>".$LocInsulinB."</font>";
      }
    }
    if (($Medicine{$_}) && ($opt_m)) {
      if ($OtherMeds) {
        $OtherMeds.="<br>";
        $OtherMedsNames.="<br>";
      }
      $OtherMeds.="<font color=purple>".$Medicine{$_}."</font>";
      if ($MedName{$_}) {
        $OtherMedsNames.="<font color=purple>".$MedName{$_}."</font>";
      }
      else {
        $OtherMedsNames.="<font color=purple>".$LocMedicine."</font>";
      }
    }
  
    $OtherMeds2="";
    $OtherMedsNames2="";
  
    if (($Carbs{$_}) && ($opt_c)) {
      $OtherMeds2.="<font color=maroon>".$Carbs{$_}."</font>";
      $OtherMedsNames2.="<font color=maroon>".$LocCarbs."</font>";
    }
    if (($Weight{$_}) && ($opt_h)) {
      if ($OtherMeds2) {
        $OtherMeds2.="<br>";
        $OtherMedsNames2.="<br>";
      }
      $OtherMeds2.="<font color=green>".$Weight{$_}."</font>";
      $OtherMedsNames2.="<font color=green>".$LocWeight."</font>";
    }
    if (($Height{$_}) && ($opt_h)) {
      if ($OtherMeds2) {
        $OtherMeds2.="<br>";
        $OtherMedsNames2.="<br>";
      }
      $OtherMeds2.="<font color=mediumvioletred>".$Height{$_}."</font>";
      $OtherMedsNames2.="<font color=mediumvioletred>".$LocHeight."</font>";
    }
    if (($BMI{$_}) && ($opt_h)) {
      if ($OtherMeds2) {
        $OtherMeds2.="<br>";
        $OtherMedsNames2.="<br>";
      }
      $OtherMeds2.="<font color=darkslategray>".$BMI{$_}."</font>";
      $OtherMedsNames2.="<font color=darkslategray>".$LocBMI."</font>";
    }
    if (($HbA1c{$_}) && ($opt_h)) {
      if ($OtherMeds2) {
        $OtherMeds2.="<br>";
        $OtherMedsNames2.="<br>";
      }
      $OtherMeds2.="<font color=blue><b>".$HbA1c{$_}."%</b></font>";
      $OtherMedsNames2.="<font color=blue><b>".$LocHbA1c."</b></font>";
    }
  
    $Ktext="";
    $Etext="";
    $Ktext="<font color=red>".$LocReportNotes1." ".$KetValue{$_}."</font><br>" if ($Ketones{$_} eq "+ve");
    $Ktext="<font color=teal>".$LocReportNotes2."</font><br>" if ($Ketones{$_} eq "-ve");
    $Etext="<font color=navy>".$LocReportNotes3.": ".$Eventname{$_}."</font><br>" if ($Eventname{$_});
    #Print end of the entry
    print "</td>";
    print "
            <td bgcolor=lightgrey width=1></td>
            <td width=35 align=right>".$OtherMeds."</td>
            <td width=5></td>
            <td width=72 style=\"color: gray;\">".$OtherMedsNames."</td>" if ($ShowOtherMed1 eq "YES");
    print "
            <td bgcolor=lightgrey width=1></td>
            <td width=30 align=right>".$OtherMeds2."</td>
            <td width=5></td>
            <td width=55 style=\"color: gray;\">".$OtherMedsNames2."</td>" if ($ShowOtherMed2 eq "YES");
    print "
            <td bgcolor=lightgrey width=1></td>
            <td style=\"color: red;\"><b>".$Notable{$_}."</b></td>
            <td width=1></td><td>".$Ktext.$Etext.$Notes{$_}."</td>
          <tr>";

    #Set values for use with next log entry
    $ThisDate=$Date{$_};
    $ThisPeriod=$Period{$_};

    #Track the Max, Min and Average glucose stats
    if ($Glucose{$_}) {
      $MaxGlucose=$Glucose{$_} if ($Glucose{$_} > $MaxGlucose);
      $MinGlucose=$Glucose{$_} if ((!$MinGlucose) || $Glucose{$_} < $MinGlucose);
      $AvgGlucose+=$Glucose{$_};
      $NumGlucose++;
    }
  }
  #All entries are now printed
  
  
  #Print the stats at the end of the report.
  $AvgGlucose=$AvgGlucose/$NumGlucose if ($AvgGlucose);
  $AvgGlucose=sprintf("%0.1f", $AvgGlucose) if ($GlucoseUnits eq "mmol/L");
  $AvgGlucose=sprintf("%0.0f", $AvgGlucose) if ($GlucoseUnits eq "mg/dL");
  
  print "
          <tr><td colspan=".$NumCols." bgcolor=".$ReportColour." height=2></td></tr>
          <tr>
          <td colspan=".$NumCols.">
            <table border=0 cellspacing=0 cellpadding=0 width=100%>
              <tr bgcolor=".$ReportColour.">
            <td align=center style=\"color: ".$HeadingColour.";\" width=10%>".$LocAverage."</td>
            <td align=center bgcolor=white width=23.3%>".$AvgGlucose."</td>
            <td align=center style=\"color: ".$HeadingColour.";\" width=10%>".$LocMinimum."</td>
            <td align=center bgcolor=white width=23.3%>".$MinGlucose."</td>
            <td align=center style=\"color: ".$HeadingColour.";\" width=10%>".$LocMaximum."</td>
            <td align=center bgcolor=white width=23.3%>".$MaxGlucose."</td>
            </tr>
          </table>
          </td>
        </tr>";
  
  #Close out the table
  print "
        </table>
      </td>
    </tr>
  </table>
  <p>&nbsp;
  ";

}


#==============================================================================
#
# SUBROUTINES - PRINT REPORTS - LOGBOOK
#
#==============================================================================

#=======================================
# Report Generation Subroutines - Log
#=======================================
sub PrintLogbookReport() {

  #Establish variables for storing glucose data
  my %MaxGlucose;
  my %MinGlucose;
  my %AvgGlucose;
  my %NumGlucose;

  #Get time ordered list of timestamps
  (@Timestamp)=sort NormalSort (keys %Glucose);

  if ($TotalPeriods > 0) {
    $NumColumns=scalar 1+(2*$TotalPeriods);
  }
  else {
    $NumColumns=1;
  }
  
  #Start report table, with Title, and column headers
  print "
  <a name=LOGBOOK></a>
  <table bgcolor=".$ReportColour."  border=0 cellspacing=0 cellpadding=2 width=683>
    <tr>
      <td>
        <table width=100% bgcolor=white border=0 cellspacing=0 cellpadding=0>
          <tr>
            <th colspan=".$NumColumns." bgcolor=".$ReportColour.">".$LocLogbookHeader."</th>
          </tr>";


  #Print each log entry from the Hash tables, extract each one by the timestamp
  undef $ThisDate;
  undef $ThisDayName;
  undef $ThisPeriod;
  undef $LastPeriod;
  undef %GlucoDay;
  undef $myDay;
  undef $myMonth;
  undef $myYear;
  undef $myLastDay;
  undef $myLastMonth;
  undef $myLastYear;
  undef $DayName;
  undef $WeekDate;
  $ThatTime=0;
  $FirstDay="START";
  foreach (@Timestamp) {

    #Save timestamp
    $Timestamp=$_;
  
    #Set first date
    $ThisDate=$Date{$Timestamp} if (!$ThisDate);
    $ThisDayName=$Day{$Timestamp} if (!$ThisDayName);
    $DayName=$Day{$Timestamp} if (!$DayName);

    $WeekDate="";
    $ThisWeek=$Date{$Timestamp} if (!$ThisWeek);
    #Print a break row if the date has changed
    if (($ThisDate ne $Date{$Timestamp}) || (($FirstDay eq "START") && ($ThisDate eq $Date{$Timestamp}))) {
      ($myDay,$myMonth,$myYear)=split $DateSep , $Date{$Timestamp} if ($DateFormat eq "day/month/year");
      ($myMonth,$myDay,$myYear)=split $DateSep , $Date{$Timestamp} if ($DateFormat eq "month/day/year");
      $ThisTime=timelocal("","","", scalar $myDay, scalar $myMonth - 1, scalar $myYear - 1900);
      $DayName=(Sun,Mon,Tue,Wed,Thu,Fri,Sat)[(localtime($ThisTime))[6]] if ($myMonth && (scalar $myMonth < 12));
      $DayName=$Day{$Timestamp};
      
      ($myLastDay,$myLastMonth,$myLastYear)=split $DateSep , $ThisWeek if ($DateFormat eq "day/month/year");
      ($myLastMonth,$myLastDay,$myLastYear)=split $DateSep , $ThisWeek if ($DateFormat eq "month/day/year");
      $ThatTime=timelocal("","","", scalar $myLastDay, scalar $myLastMonth - 1, scalar $myLastYear - 1900) if ($ThisDate);
      
      if ($DayName eq $DaysOfWeek{"Mon"}) {
        $WeekDate=$Date{$Timestamp};
        $ThisWeek=$Date{$Timestamp};
      }
      elsif (((scalar $ThisTime - scalar $ThatTime) >= 604800) || ($FirstDay eq "START")) {
        $ThatTime=scalar $ThisTime if ($DayName eq $DaysOfWeek{"Mon"});
        $ThatTime=scalar $ThisTime - (86400 * 1) if ($DayName eq $DaysOfWeek{"Tue"});
        $ThatTime=scalar $ThisTime - (86400 * 2) if ($DayName eq $DaysOfWeek{"Wed"});
        $ThatTime=scalar $ThisTime - (86400 * 3) if ($DayName eq $DaysOfWeek{"Thu"});
        $ThatTime=scalar $ThisTime - (86400 * 4) if ($DayName eq $DaysOfWeek{"Fri"});
        $ThatTime=scalar $ThisTime - (86400 * 5) if ($DayName eq $DaysOfWeek{"Sat"});
        $ThatTime=scalar $ThisTime - (86400 * 6) if ($DayName eq $DaysOfWeek{"Sun"});
        (@TWelements)=localtime($ThatTime);
        $TWDay=scalar $TWelements[3];
        $TWMonth=1 + scalar $TWelements[4];
        $TWDay="0".$TWDay if ($TWDay <= 9);
        $TWMonth="0".$TWMonth if ($TWMonth <= 9);
        $TWYear=1900 + scalar $TWelements[5];
        $ThisWeek=$TWDay.$DateSep.$TWMonth.$DateSep.$TWYear if ($DateFormat eq "day/month/year");
        $ThisWeek=$TWMonth.$DateSep.$TWDay.$DateSep.$TWYear if ($DateFormat eq "month/day/year");
        $WeekDate=$ThisWeek;
     }
      if ($FirstDay eq "START") {
        $FirstDay="YES";
        PrintLogbookEntryHeader();
      }
    }
  
    #Print previous day's details
    if ($ThisDate ne $Date{$Timestamp}) {
      PrintLogbookEntry();
      $FirstDay="NO";
    }
  
    #Determine the colour of the of the glucose field, using the preset values
    if (scalar $Glucose{$Timestamp} < $GlucoseLevels{"CriticalLow"}) {
      $Gcolor=$GlucoseColors{"VeryLow"};
    }
    elsif (scalar $Glucose{$Timestamp} < $GlucoseLevels{"TargetLow"}) {
      $Gcolor=$GlucoseColors{"Low"};
    }
    elsif (scalar $Glucose{$Timestamp} < $GlucoseLevels{"TargetHigh"}) {
      $Gcolor=$GlucoseColors{"Target"};
    }
    elsif (scalar $Glucose{$Timestamp} < $GlucoseLevels{"CriticalHigh"}) {
      $Gcolor=$GlucoseColors{"High"};
    }
    else {
      $Gcolor=$GlucoseColors{"VeryHigh"};
    }

    #Add this period's detail to the day's Hash table
    $AddNotableColour="";
    $AddNotableColour=" bgcolor=LIGHTGOLDENRODYELLOW" if ($Notable{$_});

    $GlucoDay{$Period{$Timestamp}}.="<table bgcolor=white border=0 cellspacing=0 cellpadding=0><tr><td align=right style=\"color: darkslategray;\">".$Time{$Timestamp}."</td><td>&nbsp;</td><td align=right width=25 style=\"color:".$Gcolor.";\"".$AddNotableColour.">".$Glucose{$Timestamp}."</td><td width=3>&nbsp;</td></tr></table>";

    #Track the Max, Min and Average glucose stats
    $MaxGlucose{$Period{$Timestamp}}=$Glucose{$_} if ($Glucose{$_} > $MaxGlucose{$Period{$Timestamp}});
    $MinGlucose{$Period{$Timestamp}}=$Glucose{$_} if ((!$MinGlucose{$Period{$Timestamp}}) || $Glucose{$_} < $MinGlucose{$Period{$Timestamp}});
    $AvgGlucose{$Period{$Timestamp}}+=$Glucose{$_};
    $NumGlucose{$Period{$Timestamp}}++;

    #Set date for testing if next entry is still today
    $ThisDate=$Date{$Timestamp};
    $ThisDayName=$DayName;
  }
 
  #Print last day's data 
  PrintLogbookEntry();


  #Print the average, max and min values for each period
  print "
          <tr>
            <td colspan=".$NumColumns." bgcolor=".$ReportColour." height=2></td>
          </tr>
          <tr valign=top>
            <td align=center bgcolor=".$ReportColour." style=\"color: ".$HeadingColour.";\">".$LocAverage."</td>";
  foreach (@Periods) {
  if ($AvgGlucose{$_}) {
      $AvgGlucose{$_}=$AvgGlucose{$_}/$NumGlucose{$_} if ($AvgGlucose{$_});
      $AvgGlucose{$_}=sprintf("%0.1f", $AvgGlucose{$_}) if ($GlucoseUnits eq "mmol/L");
      $AvgGlucose{$_}=sprintf("%0.0f", $AvgGlucose{$_}) if ($GlucoseUnits eq "mg/dL");
  }
  
    #Determine the colour of the of the glucose field, using the preset values
    if (scalar $AvgGlucose{$_} < $GlucoseLevels{"CriticalLow"}) {
      $Gcolor=$GlucoseColors{"VeryLow"};
    }
    elsif (scalar $AvgGlucose{$_} < $GlucoseLevels{"TargetLow"}) {
      $Gcolor=$GlucoseColors{"Low"};
    }
    elsif (scalar $AvgGlucose{$_} < $GlucoseLevels{"TargetHigh"}) {
      $Gcolor=$GlucoseColors{"Target"};
    }
    elsif (scalar $AvgGlucose{$_} < $GlucoseLevels{"CriticalHigh"}) {
      $Gcolor=$GlucoseColors{"High"};
    }
    else {
      $Gcolor=$GlucoseColors{"VeryHigh"};
    }
    print "
            <td bgcolor=".$ReportColour." width=1></td>
            <td align=right style=\"color:".$Gcolor.";\">".$AvgGlucose{$_}."&nbsp;</td>";
  }
  print "
          </tr>";

  print "
          <tr>
            <td colspan=".$NumColumns." bgcolor=".$ReportColour." height=1></td>
          </tr>
          <tr valign=top>
            <td align=center bgcolor=".$ReportColour." style=\"color: ".$HeadingColour.";\">".$LocMinimum."</td>";
  foreach (@Periods) {
 #Determine the colour of the of the glucose field, using the preset values
    if (scalar $MinGlucose{$_} < $GlucoseLevels{"CriticalLow"}) {
      $Gcolor=$GlucoseColors{"VeryLow"};
    }
    elsif (scalar $MinGlucose{$_} < $GlucoseLevels{"TargetLow"}) {
      $Gcolor=$GlucoseColors{"Low"};
    }
    elsif (scalar $MinGlucose{$_} < $GlucoseLevels{"TargetHigh"}) {
      $Gcolor=$GlucoseColors{"Target"};
    }
    elsif (scalar $MinGlucose{$_} < $GlucoseLevels{"CriticalHigh"}) {
      $Gcolor=$GlucoseColors{"High"};
    }
    else {
      $Gcolor=$GlucoseColors{"VeryHigh"};
    }
    print "
            <td bgcolor=".$ReportColour." width=1></td>
            <td align=right style=\"color:".$Gcolor.";\">".$MinGlucose{$_}."&nbsp;</td>";
  }
  print "
          </tr>";

  print "
          <tr>
            <td colspan=".$NumColumns." bgcolor=".$ReportColour." height=1></td>
          </tr>
          <tr valign=top>
            <td align=center bgcolor=".$ReportColour." style=\"color: ".$HeadingColour.";\">".$LocMaximum."</td>";
  foreach (@Periods) {
    #Determine the colour of the of the glucose field, using the preset values
    if (scalar $MaxGlucose{$_} < $GlucoseLevels{"CriticalLow"}) {
      $Gcolor=$GlucoseColors{"VeryLow"};
    }
    elsif (scalar $MaxGlucose{$_} < $GlucoseLevels{"TargetLow"}) {
      $Gcolor=$GlucoseColors{"Low"};
    }
    elsif (scalar $MaxGlucose{$_} < $GlucoseLevels{"TargetHigh"}) {
      $Gcolor=$GlucoseColors{"Target"};
    }
    elsif (scalar $MaxGlucose{$_} < $GlucoseLevels{"CriticalHigh"}) {
      $Gcolor=$GlucoseColors{"High"};
    }
    else {
      $Gcolor=$GlucoseColors{"VeryHigh"};
    }
    print "
            <td bgcolor=".$ReportColour." width=1></td>
            <td align=right style=\"color:".$Gcolor.";\">".$MaxGlucose{$_}."&nbsp;</td>";
  }
  print "
          </tr>";
  
  #Close out the table
  print "
        </table>
      </td>
    </tr>
  </table>
  <p>&nbsp;
  ";

}


#=======================================
# Report Generation Subroutines - LogDay
#=======================================
sub PrintLogbookEntry () {

  PrintLogbookEntryData();

  if ($WeekDate) {
    PrintLogbookEntryHeader();
  }
  
  #Reset Hash for next day's data
  undef %GlucoDay;

}


#=======================================
# Report Generation Subroutines - Day headr
#=======================================
sub PrintLogbookEntryHeader () {

    print "
          <tr>
            <td colspan=".$NumColumns." bgcolor=".$ReportColour." height=1></td>
          </tr>
          <tr>
            <td colspan=".$NumColumns." bgcolor=gray align=left style=\"color:".$HeadingColour.";\"><b>".$LocWeekStarting." ".$WeekDate."</b></td>
          </tr>";

    print "
          <tr>
            <td colspan=".$NumColumns." bgcolor=".$HeadingColour." height=1></td>
          </tr>
          <tr bgcolor=gray>
            <td width=50 style=\"font-size: 10px; color: ".$HeadingColour."; text-decoration:bold;\">Date</td>";

    foreach (@Periods) {
      $CurPeriod=$_;
      $CurPeriod=~ s/ /\<br\/\>/g;
      print "
         <td bgcolor=".$HeadingColour." width=1></td>
         <td align=center style=\"font-size: 10px; color: ".$HeadingColour."; text-decoration:bold;\">".$CurPeriod."</td>";
    }
        
    print "
          <tr>";
}


#=======================================
# Report Generation Subroutines - Day data
#=======================================
sub PrintLogbookEntryData () {

  ($myDay,$myMonth,$myYear)=split $DateSep , $ThisDate if ($DateFormat eq "day/month/year");
  ($myMonth,$myDay,$myYear)=split $DateSep , $ThisDate if ($DateFormat eq "month/day/year");  
  $ThisDay="<table border=0 cellpadding=0 cellspacing=0><tr><td width=25>".$ThisDayName."</td><td>(".$myDay.")</td></tr></table>";

  #Print date and then each period's data, if missing insert blank cells

  print "
    <tr>
      <td colspan=".$NumColumns." bgcolor=gray height=1></td>
    </tr>";
  print "
    <tr valign=top>
      <td>".$ThisDay."</td>";

  foreach (@Periods) {
    
    if ($GlucoDay{$_}) {
      print "
      <td bgcolor=lightgrey width=1></td>
      <td align=right>".$GlucoDay{$_}."</td>";
    }
    else {
      print "
      <td bgcolor=lightgrey width=1></td>
      <td></td>";
    }
  }
}



#==============================================================================
#
# SUBROUTINES - PRINT REPORTS - Insulin & Medicine Averages
#
#==============================================================================

#=======================================
# Report Generation Subroutines - Log
#=======================================
sub PrintInsMedAvgReport() {

  #Establish variables for storing glucose data
  my %MaxGlucose;
  my %MinGlucose;
  my %AvgGlucose;
  my %NumGlucose;

  #Get time ordered list of timestamps
  (@Timestamp)=sort NormalSort (keys %Period);

  if ($TotalPeriods > 0) {
    $NumColumns=scalar 3+(2*$TotalPeriods);
  }
  else {
    $NumColumns=3;
  }
  
  #Start report table, with Title, and column headers
  print "
  <a name=AVGLOGBOOK></a>
  <table bgcolor=".$ReportColour."  border=0 cellspacing=0 cellpadding=2 width=683>
    <tr>
  <td>
  <table width=100% bgcolor=white border=0 cellspacing=0 cellpadding=0>
  <tr>
  <th colspan=".$NumColumns." bgcolor=".$ReportColour.">".$LocAvgMedHeader."</th>
  </tr>";
  print "
  <tr>
  <td colspan=".$NumColumns." bgcolor=".$ReportColour." height=2></td>
  </tr>
  <tr>
  <td colspan=".$NumColumns." bgcolor=".$HeadingColour." height=1></td>
  </tr>
  <tr bgcolor=gray>
  <td width=50 style=\"font-size: 10px; color: ".$HeadingColour."; text-decoration:bold;\">".$LocAvgMedCol1."</td>";
  
  foreach (@Periods) {
    $CurPeriod=$_;
    $CurPeriod=~ s/ /\<br\/\>/g;
    print "
    <td bgcolor=".$HeadingColour." width=1></td>
    <td align=center style=\"font-size: 10px; color: ".$HeadingColour."; text-decoration:bold;\">".$CurPeriod."</td>";
  }
  
  print "
  <td bgcolor=".$HeadingColour." width=1></td>
  <td align=center style=\"font-size: 10px; color: ".$HeadingColour."; text-decoration:bold;\">".$LocAvgMedCol2."</td>";
  
  print "
  <tr>";
  
  
  undef $ShortTime;
  undef $MednName;
  undef %MednList;
  undef %MednListDays;
  undef %MednListPeriods;
  undef %MednListDaysCount;
  undef %MednListPeriodsCount;
  undef %MednListDaysAvg;
  undef %MednListPeriodsAvg;
  foreach (@Timestamp) {

    #Save timestamp
    $Timestamp=$_;
    $ShortTime=substr $Timestamp , 0 , 8;
    

    
    if ($Insulin{$Timestamp} && $opt_i) {
      undef @myMedList;
      undef @myMedNameList;
      undef $count;
      (@myMedList) = split "<br/>", $Insulin{$Timestamp};
      (@myMedNameList) = split "<br/>", $InsType{$Timestamp};
      for ($count = 0; $count < scalar @myMedList; $count++) {
        if ($myMedNameList[$count]) {
          $MednName=$myMedNameList[$count];
        }
        else {
          $MednName=$LocInsulinA;
        }
        $MednList{$MednName}++;
        $MednListDays{$MednName."{lbxdelim}".$ShortTime}+=$myMedList[$count];
        $MednListPeriods{$MednName."{lbxdelim}".$Period{$Timestamp}}+=$myMedList[$count];
        $MednListDaysCount{$MednName."{lbxdelim}".$ShortTime}=1;
        $MednListPeriodsCount{$MednName."{lbxdelim}".$Period{$Timestamp}}++;
      }
    }
    if ($InsulinB{$Timestamp} && $opt_i) {
      undef @myMedList;
      undef @myMedNameList;
      undef $count;
      (@myMedList) = split "<br/>", $InsulinB{$Timestamp};
      (@myMedNameList) = split "<br/>", $InsBType{$Timestamp};
      for ($count = 0; $count < scalar @myMedList; $count++) {
        if ($myMedNameList[$count]) {
          $MednName=$myMedNameList[$count];
        }
        else {
          $MednName=$LocInsulinB;
        }
        $MednList{$MednName}++;
        $MednListDays{$MednName."{lbxdelim}".$ShortTime}+=$myMedList[$count];
        $MednListPeriods{$MednName."{lbxdelim}".$Period{$Timestamp}}+=$myMedList[$count];
        $MednListDaysCount{$MednName."{lbxdelim}".$ShortTime}=1;
        $MednListPeriodsCount{$MednName."{lbxdelim}".$Period{$Timestamp}}++;
      }
    }
    if ($Medicine{$Timestamp} && $opt_m) {
      undef @myMedList;
      undef @myMedNameList;
      undef $count;
      (@myMedList) = split "<br/>", $Medicine{$Timestamp};
      (@myMedNameList) = split "<br/>", $MedName{$Timestamp};
      for ($count = 0; $count < scalar @myMedList; $count++) {
        if ($myMedNameList[$count]) {
          $MednName=$myMedNameList[$count];
        }
        else {
          $MednName=$LocMedicine;
        }
        $MednList{$MednName}++;
        $MednListDays{$MednName."{lbxdelim}".$ShortTime}+=$myMedList[$count];
        $MednListPeriods{$MednName."{lbxdelim}".$Period{$Timestamp}}+=$myMedList[$count];
        $MednListDaysCount{$MednName."{lbxdelim}".$ShortTime}=1;
        $MednListPeriodsCount{$MednName."{lbxdelim}".$Period{$Timestamp}}++;
      }
    }
  }
  
  undef %MednDayCount;
  foreach (keys %MednList) {
    $myMed=$_;
    foreach (keys %MednListDays) {
      $myMedKey=$_;
      $myMedName=(split "{lbxdelim}", $myMedKey)[0];
      if ($myMedName eq $myMed) {
        $MednListDaysAvg{$myMedName}+=$MednListDays{$myMedKey};
        $MednDayCount{$myMed}++;
      }
    }
    if ($MednListDaysAvg{$myMed}) {
      $MednListDaysAvg{$myMed}=sprintf("%0.1f", $MednListDaysAvg{$myMed} / $MednDayCount{$myMed});
      undef $MednListDaysAvg{$myMed} if $MednListDaysAvg{$myMed} == 0;
    }
  }
  foreach (keys %MednListPeriods) {
    $MednListPeriodsAvg{$_}=sprintf("%0.1f",$MednListPeriods{$_} / $MednListPeriodsCount{$_});
    undef $MednListPeriodsAvg{$_} if $MednListPeriodsAvg{$_} == 0;
  }

  undef $myMed;
  undef @myMedList;
  (@myMedList)=sort NormalSort (keys %MednList);
  
  foreach (@myMedList) {
    $myMed=$_;
    print "
    <tr>
    <td colspan=".$NumColumns." height=1 bgcolor=gray></td>
    </tr>
    <tr>
    <td>".$myMed."</td>";
    
    foreach (@Periods) {
      if ($MednListPeriodsAvg{$myMed."{lbxdelim}".$_}) {
        print "
        <td width=1 bgcolor=lightgrey></td>
        <td align=center><table width=100% border=0><tr width=100%><td align=right>".$MednListPeriodsAvg{$myMed."{lbxdelim}".$_}."</td><td width=45% align=left style=\"color: darkslategray;\">(".$MednListPeriodsCount{$myMed."{lbxdelim}".$_}.")</td></tr></table></td>";
      }
      else {
        print "
        <td width=1 bgcolor=lightgrey></td>
        <td align=right></td>";
      }
    }
    
    print "
    <td width=1 bgcolor=lightgrey></td>
    <td align=center><table width=100% border=0><tr width=100%><td align=right>".$MednListDaysAvg{$myMed}."</td><td width=45% align=left style=\"color: darkslategray;\">(".$MednDayCount{$myMed}.")</td></tr></table></td>
    </tr>";
    
  }
  
  #Close out the table
  print "
        </table>
      </td>
    </tr>
  </table>
  <p>&nbsp;
  ";

}



#==============================================================================
#
# SUBROUTINES - PRINT REPORTS - SORTING ROUTINES
#
#==============================================================================

#=======================================
# Sort function - alphabetically
#=======================================
sub NormalSort { $a cmp $b };

#=======================================
# Sort function - reverse alphabetically
#=======================================
sub InverseSort { $b cmp $a };

#=======================================
# Sort function - alphabetically
#=======================================
sub NormalSortNum { $a <=> $b };

#=======================================
# Sort function - reverse alphabetically
#=======================================
sub InverseSortNum { $b <=> $a };
